import random
from io import TextIOWrapper
from contextlib import contextmanager
from typing import Sequence, Callable, Any, Union, Optional, Tuple, TypeVar

import numpy as np
import torch
from torch import Tensor, nn


def set_seed(seed):
    seed = int(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def class_to_string(object):
    return str(object.__class__.__name__)


class CustomIterator():
    def __init__(self, data: Sequence, mapper: Callable[[Any], Any]):
        self.data = data
        self.mapper = mapper
        self.iterator = None

    def __iter__(self):
        self.iterator = iter(self.data)
        return self

    def __next__(self):
        datum = next(self.iterator)
        return self.mapper(datum)


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(
                Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Observable:
    def __init__(self):
        self._callbacks = []

    def subscribe(self, callback):
        self._callbacks.append(callback)

    def fire(self, event):
        for callback in self._callbacks:
            callback(event)


@contextmanager
def open_file_or_name(file: Union[TextIOWrapper, str], *args, **kwargs):
    file_wrapper = file
    if isinstance(file, str):
        file_wrapper = open(file, *args, **kwargs)

    yield file_wrapper

    if isinstance(file, str):
        file_wrapper.close()


def map_recursive(data, func: Callable):
    if torch.is_tensor(data):
        return func(data)
    if isinstance(data, (list, tuple, set)):
        return type(data)((map_recursive(datum, func) for datum in data))
    if isinstance(data, dict):
        return {key: map_recursive(value, func) for key, value in data.items()}
    return data


def to_device(data, device: torch.device):
    return map_recursive(data, lambda x: x.to(device))


def flatten(data):
    if isinstance(data, dict):
        data = list(data.values())
    if isinstance(data, (tuple, list)):
        return sum([flatten(elem) for elem in data], [])
    return [data]


def linear_from_params(weight: Tensor, bias: Optional[Tensor] = None) -> nn.Linear:
    layer = nn.Linear(weight.shape[1], weight.shape[0], bias=bias is not None)
    with torch.no_grad():
        layer.weight.data = weight
        layer.bias.data = bias

    return layer


T = TypeVar('T')
def format_into_tuple(obj: T) -> Tuple[T]:
    if isinstance(obj, Sequence) and not isinstance(obj, str):
        return tuple(obj)
    elif isinstance(obj, type(None)):
        return ()
    return (obj,)


def format_out_tuple_if_single(obj: Tuple[T]) -> Union[T, Tuple[T]]:
    if len(obj) == 1:
        return obj[0]
    return obj


def format_into_tuple_all(**kwargs):
    return {k: format_into_tuple(v) for k, v in kwargs.items()}


def generate_param_key(*args):
    # ensure the uniqueness of param name of optuna
    return '.'.join([str(arg) for arg in args if arg is not None])



from abc import ABC
import copy
import torch


class UtilFunction(ABC):
    def __init__(self, *args, **kwargs):
        pass

    # def __repr__(self):
    #     return "{}({!r})".format(self.__class__.__name__, self.__dict__)

    def __repr__(self):
        kwargs_repr = ', '.join('{}={}'.format(key, value) for key, value in self.__dict__.items())
        return "{}({})".format(self.__class__.__name__, kwargs_repr)

    def copy(self):
        return copy.copy(self)

    def set_kwargs(self, **kwargs):
        clone = self.copy()
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(clone, k, v)
        return clone

    def __call__(self, inputs: torch.Tensor):
        return NotImplementedError

    def get_tunables(self):
        return {}



from pnpxai.explainers.rap.rules.base import RelProp

from pnpxai.explainers.rap.rules.functions import (
    ReLU, GeLU, LeakyReLU, SoftMax,
    Add, Sub, Mul, Div, FloorDiv, MatMul,
    Flatten, Cat, Repeat,
    Squeeze, Unsqueeze, Expand,
    GetItem, GetAttr,
    Permute, Reshape, Transpose, View, Contiguous, Unfold, ReplicationPad1d,
)

from pnpxai.explainers.rap.rules.modules import (
    Dropout,
    MaxPool2d, MaxPool1d,
    AdaptiveAvgPool2d, AdaptiveAvgPool1d,
    AvgPool1d, AvgPool2d,
    BatchNorm2d, BatchNorm1d,
    LayerNorm,
    Linear,
    Conv2d, Conv1d,
    MultiHeadAttention
)


import _operator
from typing import Dict, Type
import torch
from torch import nn
from torch.nn import functional as F
from pnpxai.explainers.rap import rules

SUPPORTED_OPS: Dict[str, Dict[str, Type[rules.RelProp]]] = {
    'call_module': {
        nn.ReLU: rules.ReLU,
        nn.GELU: rules.GeLU,
        nn.LeakyReLU: rules.LeakyReLU,
        nn.Dropout: rules.Dropout,
        nn.MaxPool2d: rules.MaxPool2d,
        nn.MaxPool1d: rules.MaxPool1d,
        nn.AdaptiveAvgPool2d: rules.AdaptiveAvgPool2d,
        nn.AdaptiveAvgPool1d: rules.AdaptiveAvgPool1d,
        nn.AvgPool2d: rules.AvgPool2d,
        nn.AvgPool1d: rules.AvgPool1d,
        nn.BatchNorm2d: rules.BatchNorm2d,
        nn.BatchNorm1d: rules.BatchNorm1d,
        nn.LayerNorm: rules.LayerNorm,
        nn.Linear: rules.Linear,
        nn.Conv2d: rules.Conv2d,
        nn.Conv1d: rules.Conv1d,
        nn.MultiheadAttention: rules.MultiHeadAttention,
        nn.Flatten: rules.Flatten,
        nn.Softmax: rules.SoftMax,
        nn.Unfold: rules.Unfold,
        nn.ReplicationPad1d: rules.ReplicationPad1d,
    },
    'call_function': {
        getattr: rules.GetAttr,
        _operator.add: rules.Add,
        _operator.sub: rules.Sub,
        _operator.mul: rules.Mul,
        _operator.matmul: rules.MatMul,
        _operator.truediv : rules.Div,
        _operator.floordiv: rules.FloorDiv,
        _operator.getitem: rules.GetItem,
        torch.add: rules.Add,
        torch.mul: rules.Mul,
        torch.div: rules.Div,
        torch.matmul: rules.MatMul,
        torch.flatten: rules.Flatten,
        torch.relu: rules.ReLU,
        torch.softmax: rules.SoftMax,
        torch.cat: rules.Cat,
        torch.transpose: rules.Flatten,
        torch.sub: rules.Sub,
        torch.squeeze: rules.Squeeze,
        torch.unsqueeze: rules.Unsqueeze,
        F.avg_pool2d: rules.AvgPool2d,
        F.avg_pool1d: rules.AvgPool1d,
        F.max_pool2d: rules.MaxPool2d,
        F.max_pool1d: rules.MaxPool1d,
        F.relu: rules.ReLU,
        F.gelu: rules.GeLU,
        F.leaky_relu: rules.LeakyReLU,
        F.softmax: rules.SoftMax,
        F.unfold: rules.Unfold,
    },
    'call_method': {
        'add': rules.Add,
        'sub': rules.Sub,
        'mul': rules.Mul,
        'div': rules.Div,
        'matmul': rules.MatMul,
        'relu': rules.ReLU,
        'softmax': rules.SoftMax,
        'transpose': rules.Flatten,
        'repeat': rules.Repeat,
        'expand': rules.Expand,
        'permute': rules.Permute,
        'reshape': rules.Reshape,
        'squeeze': rules.Squeeze,
        'view': rules.View,
        'contiguous': rules.Contiguous,
        'unfold': rules.Unfold,
    }

}


'''Additional Utility Layers'''
import torch
from zennit.layer import Sum

class StackAndSum(torch.nn.Module):
    '''Compute the sum along an axis.

    Parameters
    ----------
    dim : int
        Dimension over which to sum.
    '''
    def __init__(self, dim=-1):
        super().__init__()
        self.dim = dim
        self.zennit_sum = Sum(dim=dim)

    def forward(self, a, b):
        out = torch.stack([a, b.expand(a.size())], dim=self.dim)
        '''Computes the sum along a dimension.'''
        return self.zennit_sum(out)



import torch
import torch.nn.functional as F
from zennit.core import Hook, Stabilizer

from pnpxai.explainers.zennit.hooks import HookWithKwargs


class AttentionHeadRule(HookWithKwargs):
    def __init__(
            self,
            stabilizer=1e-6,
        ):
        super().__init__()
        self.stabilizer = stabilizer

    def backward(self, module, grad_input, grad_output):
        query, key, value = (
            inp.clone().requires_grad_()
            for inp in self.stored_tensors["input"]
        )
        key_padding_mask, need_weights, attn_mask, average_attn_weights, is_causal = self._parse_kwargs(self.stored_kwargs)

        with torch.autograd.enable_grad():
            # Preprocess
            is_batched = F._mha_shape_check(query, key, value, key_padding_mask, attn_mask, module.num_heads)
            if module.batch_first and is_batched:
                query, key, value = (x.transpose(1, 0) for x in (query, key, value))
            if not is_batched:
                query = query.unsqueeze(1)
                key = key.unsqueeze(1)
                value = value.unsqueeze(1)
                if key_padding_mask is not None:
                    key_padding_mask = key_padding_mask.unsqueeze(0)

            key_padding_mask = F._canonical_mask(
                mask=key_padding_mask,
                mask_name="key_padding_mask",
                other_type=F._none_or_dtype(attn_mask),
                other_name="attn_mask",
                target_type=query.dtype
            )

            if is_causal and key_padding_mask is None and not need_weights:
                attn_mask = None
            else:
                attn_mask = F._canonical_mask(
                    mask=attn_mask,
                    mask_name="attn_mask",
                    other_type=None,
                    other_name="",
                    target_type=query.dtype,
                    check_other=False,
                )

                if key_padding_mask is not None:
                    is_causal = False
            # Forward
            attn_output_weights = self._attention_weights_forward(module, query, key, module.bias_k, key_padding_mask, attn_mask)

            in_proj_intermediate = self._in_proj_intermediate_forward(module, value, attn_output_weights)
            in_proj_output = self._in_proj_output_forward(module, in_proj_intermediate, attn_output_weights)

            attn_output = self._out_projection_forward(module, in_proj_output, query.shape[0], query.shape[1])

        stabilizer_fn = Stabilizer.ensure(self.stabilizer)

        # Relevance: in_proj_output
        grad_outputs = grad_output[0] / stabilizer_fn(attn_output)
        gradients = torch.autograd.grad(
            attn_output,
            in_proj_output,
            grad_outputs=grad_outputs,
            create_graph=grad_output[0].requires_grad,
        )
        relevance_in_proj_output = in_proj_output * gradients[0]

        # Relevance: in_proj_intermediate
        grad_outputs = relevance_in_proj_output[0] / stabilizer_fn(in_proj_output)
        gradients = torch.autograd.grad(
            in_proj_output,
            in_proj_intermediate,
            grad_outputs=grad_outputs,
        )
        relevance_in_proj_intermediate = in_proj_intermediate * gradients[0]

        # Relevance: value  
        grad_outputs = relevance_in_proj_intermediate[0] / stabilizer_fn(in_proj_intermediate)
        gradients = torch.autograd.grad(
            in_proj_intermediate,
            value,
            grad_outputs=grad_outputs,
        )
        relevance_value = value * gradients[0]

        if module.batch_first and is_batched:
            query, key, relevance_value = (x.transpose(1, 0) for x in (query, key, relevance_value))

        return (
            torch.zeros_like(query),
            torch.zeros_like(key),
            relevance_value,
        )

    def _parse_kwargs(self, kwargs):
        key_padding_mask = kwargs.get('key_padding_mask')
        need_weights = kwargs.get('need_weights')
        attn_mask = kwargs.get('attn_mask')
        average_attn_weights = kwargs.get('average_attn_weights')
        is_causal = kwargs.get('is_causal')
        return key_padding_mask, need_weights, attn_mask, average_attn_weights, is_causal

    def copy(self):
        copy = AttentionHeadRule.__new__(type(self))
        AttentionHeadRule.__init__(
            copy,
            self.stabilizer
        )
        return copy

    @staticmethod
    def _attention_weights_forward(module, query, key, bias_k, key_padding_mask, attn_mask):
        tgt_len, bsz, embed_dim = query.shape
        src_len, _, _ = key.shape
        num_heads, head_dim = module.num_heads, module.head_dim

        if module._qkv_same_embed_dim:
            indices = [(i+1)*module.kdim for i in range(2)]
            w_q = module.in_proj_weight[:indices[0]]
            w_k = module.in_proj_weight[indices[0]:indices[1]]
        else:
            w_q, w_k = module.q_proj_weight, module.k_proj_weight

        if module.in_proj_bias is not None:
            b_q, b_k, _ = module.in_proj_bias.chunk(3)
        else:
            b_q = b_k = None

        # in projection
        q = F.linear(query, w_q, b_q)
        k = F.linear(key, w_k, b_k)

        # add bias along batch dimension
        if bias_k is not None:
            k = torch.cat([k, bias_k.repeat(1, bsz, 1)])
            if attn_mask is not None:
                attn_mask = F.pad(attn_mask, (0, 1))
            if key_padding_mask is not None:
                key_padding_mask = F.pad(key_padding_mask, (0, 1))
        else:
            assert bias_k is None
        q = q.view(tgt_len, bsz * num_heads, head_dim).transpose(0, 1)
        k = k.view(src_len, bsz * num_heads, head_dim).transpose(0, 1)

        # add zero attention along batch dimension (now first)
        if module.add_zero_attn:
            zero_attn_shape = (bsz * num_heads, 1, head_dim)
            k = torch.cat([k, torch.zeros(zero_attn_shape, dtype=k.dtype, device=k.device)], dim=1)
            if attn_mask is not None:
                attn_mask = F.pad(attn_mask, (0, 1))
            if key_padding_mask is not None:
                key_padding_mask = F.pad(key_padding_mask, (0, 1))

        # update source sequence length after adjustments
        src_len = k.size(1)

        # merge key padding and attention masks
        if key_padding_mask is not None:
            assert key_padding_mask.shape == (bsz, src_len), \
                f"expecting key_padding_mask shape of {(bsz, src_len)}, but got {key_padding_mask.shape}"
            key_padding_mask = key_padding_mask.view(bsz, 1, 1, src_len).   \
                expand(-1, num_heads, -1, -1).reshape(bsz * num_heads, 1, src_len)
            if attn_mask is None:
                attn_mask = key_padding_mask
            else:
                attn_mask = attn_mask + key_padding_mask

        # calculate attention
        B, Nt, E = q.shape
        q_scaled = q / (E**.5)
        if attn_mask is not None:
            attn_output_weights = torch.baddbmm(attn_mask, q_scaled, k.transpose(-2, -1))
        else:
            attn_output_weights = torch.bmm(q_scaled, k.transpose(-2, -1))
        attn_output_weights = F.softmax(attn_output_weights, dim=-1)

        return attn_output_weights

    @staticmethod
    def _in_proj_intermediate_forward(module, value, attn_output_weights):
        src_len, bsz, embed_dim = value.shape
        num_heads, head_dim = module.num_heads, module.head_dim
        tgt_len = attn_output_weights.shape[1]

        # (bsz * num_heads, tgt_len, src_len) > (bsz, num_heads * tgt_len, src_len)
        attn_output_weights = attn_output_weights.view(bsz, num_heads * tgt_len, src_len)
        
        # (src_len, bsz, embed_dim) > (bsz, src_len, embed_dim)
        value = value.transpose(0, 1)

        # (bsz, num_heads * tgt_len, embed_dim)
        in_proj_intermediate = torch.bmm(attn_output_weights, value)

        # (bsz, num_heads, tgt_len, embed_dim)
        in_proj_intermediate = in_proj_intermediate.view(bsz, num_heads, tgt_len, embed_dim)

        # (bsz * num_heads, tgt_len, embed_dim)
        in_proj_intermediate = in_proj_intermediate.view(bsz * num_heads, tgt_len, embed_dim)

        return in_proj_intermediate

    @staticmethod
    def _in_proj_output_forward(module, in_proj_intermediate, attn_output_weights):
        if module._qkv_same_embed_dim:
            w_v = module.in_proj_weight[2*module.kdim:3*module.kdim]
        else:
            w_v = module.v_proj_weight

        if module.in_proj_bias is not None:
            _, _, b_v = module.in_proj_bias.chunk(3)
        else:
            b_v = None
        num_heads, head_dim = module.num_heads, module.head_dim
        embed_dim = w_v.shape[0]
        bsz_num_heads, tgt_len, src_len = attn_output_weights.shape
        bsz = bsz_num_heads // num_heads

        # (num_heads, bsz * tgt_len, embed_dim)
        in_proj_intermediate = in_proj_intermediate.view(bsz, num_heads, tgt_len, embed_dim).transpose(0, 1).contiguous().view(num_heads, bsz * tgt_len, embed_dim)
        
        # (num_heads, embed_dim, head_dim)
        w_v = w_v.T.view(embed_dim, num_heads, head_dim).transpose(0, 1)
        
        # (num_heads, bsz * tgt_len, head_dim)
        in_proj_output = torch.bmm(in_proj_intermediate, w_v)
        
        # (tgt_len, bsz, num_heads, head_dim)
        in_proj_output = in_proj_output.view(num_heads, bsz, tgt_len, head_dim).permute(2, 1, 0, 3)

        if module.in_proj_bias is not None:
            # (bsz * num_heads, src_len, head_dim)
            b_v = torch.tile(b_v, (src_len, bsz, 1)).view(src_len, bsz * num_heads, head_dim).transpose(0, 1)

            # (tgt_len, bsz, num_heads, head_dim)
            scaled_b_v = torch.bmm(attn_output_weights, b_v).view(bsz, num_heads, tgt_len, head_dim).permute(2, 0, 1, 3)

            # (tgt_len, bsz, num_heads, head_dim)
            in_proj_output += scaled_b_v

        # (tgt_len * bsz, embed_dim)
        in_proj_output = in_proj_output.contiguous().view(tgt_len * bsz, embed_dim)

        return in_proj_output

    @staticmethod
    def _out_projection_forward(module, in_proj_output, tgt_len, bsz):
        attn_output = F.linear(in_proj_output, module.out_proj.weight, module.out_proj.bias)
        attn_output = attn_output.view(tgt_len, bsz, attn_output.size(1)).transpose(0, 1)
        return attn_output


class LayerNormRule(Hook):
    def __init__(self, stabilizer=1e-6):
        super().__init__()
        self.stabilizer = stabilizer

    def forward(self, module, input, output):
        '''Forward hook to save module in-/outputs.'''
        self.stored_tensors['input'] = input

    def backward(self, module, grad_input, grad_output):
        input = self.stored_tensors["input"][0].clone().requires_grad_()
        with torch.autograd.enable_grad():
            # Preprocess
            dims = tuple(range(-len(module.normalized_shape), -2, -1))
            
            # Normalization: centering
            mean = input.mean(dims, keepdims=True)
            centered = input - mean

            # Normalization: rescaling
            # correction should be 1 as it is calculated via the unbiased estimator.
            var = input.var(dims, keepdims=True, correction=1).detach()
            rescaled = centered / (var + module.eps).sqrt()

            # Elementwise affine transformation
            transformed = torch.mul(rescaled, module.weight) + module.bias

        stabilizer_fn = Stabilizer.ensure(self.stabilizer)

        ''' # Separate relevance propagation including affine transformation
        # Relevance: rescaled
        grad_outputs = grad_output[0] / stabilizer_fn(transformed)
        gradients = torch.autograd.grad(
            transformed,
            rescaled,
            grad_outputs=grad_outputs,
            create_graph=grad_output[0].requires_grad,
        )
        relevance_rescaled = rescaled * gradients[0]

        # Relevance: input
        # pdb.set_trace()
        grad_outputs = relevance_rescaled[0] / stabilizer_fn(rescaled)
        gradients = torch.autograd.grad(
            rescaled,
            input,
            grad_outputs=grad_outputs,
        )
        relevance_input = input * gradients[0]
        '''

        # ''' # End-to-end relevance propagation including affine transformation
        # Relevance: end-to-end
        grad_outputs = grad_output[0] / stabilizer_fn(transformed)
        gradients = torch.autograd.grad(
            transformed,
            input,
            grad_outputs=grad_outputs,
            create_graph=grad_output[0].requires_grad,
        )
        relevance_input = input * gradients[0]
        # '''

        ''' # Relevance: end-to-end without affine transformation
        grad_outputs = grad_output[0] / stabilizer_fn(rescaled)
        gradients = torch.autograd.grad(
            rescaled,
            input,
            grad_outputs=grad_outputs,
            create_graph=grad_output[0].requires_grad,
        )
        relevance_input = input * gradients[0]
        '''
        return tuple(relevance_input if original.shape == relevance_input.shape else None for original in grad_input)


from pnpxai.explainers.rap.rules.base import (
    RelProp,
    RelPropSimple,
    _TensorOrTensors,
    safe_divide,
)
from torch import Tensor
import torch


class ReLU(RelProp):
    pass


class GeLU(RelProp):
    pass


class LeakyReLU(RelProp):
    pass


class SoftMax(RelProp):
    def relprop(
        self,
        rel: _TensorOrTensors,
        inputs: _TensorOrTensors,
        outputs: _TensorOrTensors,
        args=None,
        kwargs=None,
    ):
        dim = kwargs.get("dim", None)
        
        if dim is None:
            # if args[0] is tensor, then dim = args[1]
            if torch.is_tensor(args[0]) and len(args) > 1:
                dim = args[1]
            elif isinstance(args[0], int):
                dim = args[0]


        rel = (rel - (outputs * rel.sum(dim, keepdim=True))) * inputs
        return rel


class Add(RelPropSimple):
    def _partial_relprop(
        self,
        rel: Tensor,
        inputs: _TensorOrTensors,
        is_pos: bool,
        args=None,
        kwargs=None,
    ):
        clamp_kwargs = {"min": 0} if is_pos else {"max": 0}
        inputs = [torch.clamp(x, **clamp_kwargs) for x in inputs]
        pos_outputs = torch.add(*inputs)
        return super().relprop(rel, inputs, pos_outputs, args, kwargs)

    def relprop(
        self,
        rel: _TensorOrTensors,
        inputs: _TensorOrTensors,
        outputs: _TensorOrTensors,
        args=None,
        kwargs=None,
    ):
        if torch.is_tensor(inputs):
            return rel

        pos_rel = self._partial_relprop(rel, inputs, True, args, kwargs)
        neg_rel = self._partial_relprop(rel, inputs, False, args, kwargs)

        if torch.is_tensor(pos_rel):
            return pos_rel + neg_rel
        return [p + n for p, n in zip(pos_rel, neg_rel)]


class Sub(Add):
    pass


class Mul(RelPropSimple):
    def relprop(
        self,
        rel: _TensorOrTensors,
        inputs: _TensorOrTensors,
        outputs: _TensorOrTensors,
        args=None,
        kwargs=None,
    ):
        if torch.is_tensor(inputs):
            inputs = [inputs]

        inputs = [val for val in inputs if torch.is_tensor(val)]
        if len(inputs) <= 1:
            return rel

        return super().relprop(rel, inputs, outputs, args, kwargs)


class FloorDiv(Mul):
    pass


class Div(Mul):
    pass


class MatMul(RelProp):
    def relprop(
        self,
        rel: _TensorOrTensors,
        inputs: _TensorOrTensors,
        outputs: _TensorOrTensors,
        args=None,
        kwargs=None,
    ) -> _TensorOrTensors:
        in_a, in_b = inputs
        rel_norm = safe_divide(rel, outputs)

        rel_a = (rel_norm @ in_b.transpose(-1, -2)) * in_a
        rel_b = (in_a.transpose(-1, -2) @ rel_norm) * in_b

        return rel_a, rel_b


class Flatten(RelProp):
    def relprop(
        self,
        rel: _TensorOrTensors,
        inputs: _TensorOrTensors,
        outputs: _TensorOrTensors,
        args=None,
        kwargs=None,
    ):
        return rel.reshape(inputs.shape)


class Cat(RelPropSimple):
    def backward(
        self,
        rel: _TensorOrTensors,
        inputs: _TensorOrTensors,
        outputs: _TensorOrTensors,
        args=None,
        kwargs=None,
    ):
        Sp = safe_divide(rel, outputs)
        Cp = self.gradprop(outputs, inputs, Sp)
        rel = [x * cp for x, cp in zip(inputs, Cp)]

        return rel


class Repeat(RelPropSimple):
    pass


class GetItem(RelPropSimple):
    def relprop(
        self, rel: Tensor, inputs: Tensor, outputs: Tensor, args=None, kwargs=None
    ):
        # If module's output is a tuple, propagate rel as is
        inputs = args[0]
        if not torch.is_tensor(inputs):
            return rel

        rel_fill = torch.zeros_like(inputs)
        rel_fill[args[1]] = rel

        return rel_fill


class Unsqueeze(RelProp):
    def relprop(
        self, rel: Tensor, inputs: Tensor, outputs: Tensor, args=None, kwargs=None
    ) -> _TensorOrTensors:
        if "input" in kwargs:
            del kwargs["input"]
        else:
            args = args[1:]

        rel = torch.squeeze(rel, *args, **kwargs)
        return rel


class Squeeze(RelPropSimple):
    pass


class Expand(RelPropSimple):
    pass


class Permute(RelPropSimple):
    def relprop(
        self, rel: Tensor, inputs: Tensor, outputs: Tensor, args=None, kwargs=None
    ) -> _TensorOrTensors:
        dims = kwargs.get("dims", None)
        if dims is None:
            dims = args[1] if isinstance(args[1], (tuple, list)) else args[1:]

        dims = torch.LongTensor(dims)
        inv = torch.empty_like(dims)
        inv[dims] = torch.arange(len(dims), device=dims.device)

        return rel.permute(inv.tolist())


class Reshape(RelProp):
    def relprop(
        self, rel: Tensor, inputs: Tensor, outputs: Tensor, args=None, kwargs=None
    ) -> _TensorOrTensors:
        return rel.reshape(inputs.shape)


class Transpose(RelProp):
    def relprop(
        self, rel: Tensor, inputs: Tensor, outputs: Tensor, args=None, kwargs=None
    ) -> Tensor:
        dim1 = kwargs.get("dim1", args[1] if len(args) > 1 else None)
        dim2 = kwargs.get("dim2", args[2] if len(args) > 2 else None)

        return rel.transpose(dim1, dim2)


class View(RelProp):
    def relprop(
        self, rel: Tensor, inputs: Tensor, outputs: Tensor, args=None, kwargs=None
    ) -> Tensor:
        return rel.contiguous().view_as(inputs)


class GetAttr(RelProp):
    pass


class Contiguous(RelProp):
    pass


class Unfold(RelPropSimple):
    pass


class ReplicationPad1d(RelPropSimple):
    pass



from typing import Optional, List, Callable, Type, Any, Dict
from abc import ABC, abstractmethod

from pnpxai.explainers.utils import UtilFunction
from pnpxai.explainers.utils.postprocess import (
    PostProcessor,
    POOLING_FUNCTIONS,
    POOLING_FUNCTIONS_FOR_IMAGE,
    POOLING_FUNCTIONS_FOR_TEXT,
    POOLING_FUNCTIONS_FOR_TIME_SERIES,
    NORMALIZATION_FUNCTIONS_FOR_IMAGE,
    NORMALIZATION_FUNCTIONS_FOR_TEXT,
    NORMALIZATION_FUNCTIONS_FOR_TIME_SERIES,
    PoolingFunction,
    NormalizationFunction
)
from pnpxai.explainers.utils.function_selectors import FunctionSelector
from pnpxai.explainers.utils.baselines import (
    BaselineFunction,
    BASELINE_FUNCTIONS,
    BASELINE_FUNCTIONS_FOR_IMAGE,
    BASELINE_FUNCTIONS_FOR_TEXT,
    BASELINE_FUNCTIONS_FOR_TIME_SERIES,
)
from pnpxai.explainers.utils.feature_masks import (
    FeatureMaskFunction,
    FEATURE_MASK_FUNCTIONS,
    FEATURE_MASK_FUNCTIONS_FOR_IMAGE,
    FEATURE_MASK_FUNCTIONS_FOR_TEXT,
    FEATURE_MASK_FUNCTIONS_FOR_TIME_SERIES,
)
from pnpxai.explainers import (
    Gradient,
    GradientXInput,
    SmoothGrad,
    VarGrad,
    IntegratedGradients,
    LRPUniformEpsilon,
    LRPEpsilonPlus,
    LRPEpsilonGammaBox,
    LRPEpsilonAlpha2Beta1,
    KernelShap,
    Lime,
    AVAILABLE_EXPLAINERS
)


class Modality(ABC):
    """
    An abstract class describing modality-specific workflow. The class is used to define both default and available
    explainers, baselines, feature masks, pooling methods, and normalization methods for the modality.

    Parameters:
        channel_dim (int): Target sequence dimension.
        baseline_fn_selector (Optional[FunctionSelector]): Selector of baselines for the modality's explainers. If None selected, all BASELINE_FUNCTIONS will be used.
        feature_mask_fn_selector (Optional[FunctionSelector]): Selector of feature masks for the modality's explainers. If None selected, all FEATURE_MASK_FUNCTIONS will be used.
        pooling_fn_selector (Optional[FunctionSelector]): Selector of pooling methods for the modality's explainers. If None selected, all POOLING_FUNCTIONS will be used.
        normalization_fn_selector (Optional[FunctionSelector]): Selector of normalization methods for the modality's explainers. If None selected, all NORMALIZATION_FUNCTIONS_FOR_IMAGE will be used.

    Attributes:
        EXPLAINERS (Tuple[Explainer]): Tuple of all available explainers.
    """

    # Copies the tuple without preserving the reference
    EXPLAINERS = tuple(iter(AVAILABLE_EXPLAINERS))

    def __init__(
        self,
        channel_dim: int,
        baseline_fn_selector: Optional[FunctionSelector] = None,
        feature_mask_fn_selector: Optional[FunctionSelector] = None,
        pooling_fn_selector: Optional[FunctionSelector] = None,
        normalization_fn_selector: Optional[FunctionSelector] = None,
        **kwargs
    ):
        self.channel_dim = channel_dim
        self.baseline_fn_selector = baseline_fn_selector or FunctionSelector(BASELINE_FUNCTIONS)
        self.feature_mask_fn_selector = feature_mask_fn_selector or FunctionSelector(FEATURE_MASK_FUNCTIONS)
        self.pooling_fn_selector = pooling_fn_selector or FunctionSelector(POOLING_FUNCTIONS)
        self.normalization_fn_selector = normalization_fn_selector or FunctionSelector(NORMALIZATION_FUNCTIONS_FOR_IMAGE)

    @abstractmethod    
    def get_default_feature_mask_fn(self) -> Callable:
        """
        Defines default baseline function for the modality's explainers.

        Returns:
            BaselineFunction: Zeros baseline function.
        """
        raise NotImplementedError

    @abstractmethod
    def get_default_baseline_fn(self) -> Callable:
        """
        Defines default feature mask function for the modality's explainers.

        Returns:
            FeatureMaskFunction: No Mask baseline function.
        """
        raise NotImplementedError

    @abstractmethod
    def get_default_postprocessors(self) -> List[Callable]:
        """
        Defines default post-processors list for the modality's explainers.

        Returns:
            List[PostProcessor]: Identity PostProcessors.
        """
        raise NotImplementedError
    
    def map_fn_selector(self, method_type: Type[Any]) -> Dict[Type[UtilFunction], callable]:
        """
        Selects custom optimizable hyperparameter functions.

        Returns:
            Dict[Type[UtilFunction], callable]: Identity PostProcessors.
        """
        return {
            BaselineFunction: self.baseline_fn_selector,
            FeatureMaskFunction: self.feature_mask_fn_selector,
            PoolingFunction: self.pooling_fn_selector,
            NormalizationFunction: self.normalization_fn_selector,
        }.get(method_type, None)


class ImageModality(Modality):
    """
    An extension of Modality class for Image domain with automatic explainers and evaluation metrics recommendation.

    Parameters:
        channel_dim (int): Target sequence dimension.
        baseline_fn_selector (Optional[FunctionSelector]): Selector of baselines for the modality's explainers. If None selected, BASELINE_FUNCTIONS_FOR_TIME_SERIES will be used.
        feature_mask_fn_selector (Optional[FunctionSelector]): Selector of feature masks for the modality's explainers. If None selected, FEATURE_MASK_FUNCTIONS_FOR_TIME_SERIES will be used.
        pooling_fn_selector (Optional[FunctionSelector]): Selector of pooling methods for the modality's explainers. If None selected, POOLING_FUNCTIONS_FOR_TIME_SERIES will be used.
        normalization_fn_selector (Optional[FunctionSelector]): Selector of normalization methods for the modality's explainers. If None selected, NORMALIZATION_FUNCTIONS_FOR_TIME_SERIES will be used.
    """
    def __init__(
        self,
        channel_dim: int = 1,
        baseline_fn_selector: Optional[FunctionSelector] = None,
        feature_mask_fn_selector: Optional[FunctionSelector] = None,
        pooling_fn_selector: Optional[FunctionSelector] = None,
        normalization_fn_selector: Optional[FunctionSelector] = None,
    ):
        super(ImageModality, self).__init__(
            channel_dim,
            baseline_fn_selector=baseline_fn_selector or FunctionSelector(
                data=BASELINE_FUNCTIONS_FOR_IMAGE,
                default_kwargs={'dim': channel_dim},
            ),
            feature_mask_fn_selector=feature_mask_fn_selector or FunctionSelector(
                data=FEATURE_MASK_FUNCTIONS_FOR_IMAGE
            ),
            pooling_fn_selector=pooling_fn_selector or FunctionSelector(
                data=POOLING_FUNCTIONS_FOR_IMAGE,
                default_kwargs={'channel_dim': channel_dim},
            ),
            normalization_fn_selector=normalization_fn_selector or FunctionSelector(
                data=NORMALIZATION_FUNCTIONS_FOR_IMAGE
            ),
        )

    def get_default_baseline_fn(self) -> BaselineFunction:
        """
        Defines default baseline function for the modality's explainers.

        Returns:
            BaselineFunction: Zeros baseline function.
        """
        return self.baseline_fn_selector.select('zeros')

    def get_default_feature_mask_fn(self) -> FeatureMaskFunction:
        """
        Defines default feature mask function for the modality's explainers.

        Returns:
            FeatureMaskFunction: Felzenszwalb baseline function.
        """
        return self.feature_mask_fn_selector.select('felzenszwalb', scale=250)

    def get_default_postprocessors(self) -> List[PostProcessor]:
        """
        Defines default post-processors list for the modality's explainers.

        Returns:
            List[PostProcessor]: All available PostProcessors.
        """
        return [
            PostProcessor(
                pooling_fn=self.pooling_fn_selector.select(pm),
                normalization_fn=self.normalization_fn_selector.select(nm),
            ) for pm in self.pooling_fn_selector.choices
            for nm in self.normalization_fn_selector.choices
        ]


class TextModality(Modality):
    """
    An extension of Modality class for Text domain with automatic explainers and evaluation metrics recommendation.

    Parameters:
        channel_dim (int): Target sequence dimension.
        baseline_fn_selector (Optional[FunctionSelector]): Selector of baselines for the modality's explainers. If None selected, BASELINE_FUNCTIONS_FOR_TEXT will be used.
        feature_mask_fn_selector (Optional[FunctionSelector]): Selector of feature masks for the modality's explainers. If None selected, FEATURE_MASK_FUNCTIONS_FOR_TEXT will be used.
        pooling_fn_selector (Optional[FunctionSelector]): Selector of pooling methods for the modality's explainers. If None selected, POOLING_FUNCTIONS_FOR_TEXT will be used.
        normalization_fn_selector (Optional[FunctionSelector]): Selector of normalization methods for the modality's explainers. If None selected, NORMALIZATION_FUNCTIONS_FOR_TEXT will be used.
    """
    EXPLAINERS = (
        Gradient,
        GradientXInput,
        SmoothGrad,
        VarGrad,
        IntegratedGradients,
        LRPUniformEpsilon,
        LRPEpsilonPlus,
        LRPEpsilonGammaBox,
        LRPEpsilonAlpha2Beta1,
        KernelShap,
        Lime,
    )

    def __init__(
        self,
        channel_dim: int = -1,
        mask_token_id: int = 0,
        baseline_fn_selector: Optional[FunctionSelector] = None,
        feature_mask_fn_selector: Optional[FunctionSelector] = None,
        pooling_fn_selector: Optional[FunctionSelector] = None,
        normalization_fn_selector: Optional[FunctionSelector] = None,
    ):
        super(TextModality, self).__init__(
            channel_dim,
            baseline_fn_selector=baseline_fn_selector or FunctionSelector(
                data=BASELINE_FUNCTIONS_FOR_TEXT,
                default_kwargs={'token_id': mask_token_id},
            ),
            feature_mask_fn_selector=feature_mask_fn_selector or FunctionSelector(
                data=FEATURE_MASK_FUNCTIONS_FOR_TEXT
            ),
            pooling_fn_selector=pooling_fn_selector or FunctionSelector(
                data=POOLING_FUNCTIONS_FOR_TEXT,
                default_kwargs={'channel_dim': channel_dim},
            ),
            normalization_fn_selector=normalization_fn_selector or FunctionSelector(
                data=NORMALIZATION_FUNCTIONS_FOR_TEXT,
            ),
        )
        self.mask_token_id = mask_token_id

    def get_default_baseline_fn(self) -> BaselineFunction:
        """
        Defines default baseline function for the modality's explainers.

        Returns:
            BaselineFunction: Token baseline function.
        """
        return self.baseline_fn_selector.select('token')

    def get_default_feature_mask_fn(self) -> FeatureMaskFunction:
        """
        Defines default feature mask function for the modality's explainers.

        Returns:
            FeatureMaskFunction: No Mask baseline function.
        """
        return self.feature_mask_fn_selector.select('no_mask_1d')

    def get_default_postprocessors(self) -> List[PostProcessor]:
        """
        Defines default post-processors list for the modality's explainers.

        Returns:
            List[PostProcessor]: All PostProcessors.
        """
        return [
            PostProcessor(
                pooling_fn=self.pooling_fn_selector.select(pm),
                normalization_fn=self.normalization_fn_selector.select(nm),
            ) for pm in self.pooling_fn_selector.choices
            for nm in self.normalization_fn_selector.choices
        ]


class TimeSeriesModality(Modality):
    """
    An extension of Modality class for Time Series domain with automatic explainers and evaluation metrics recommendation.

    Parameters:
        channel_dim (int): Target sequence dimension.
        baseline_fn_selector (Optional[FunctionSelector]): Selector of baselines for the modality's explainers. If None selected, BASELINE_FUNCTIONS_FOR_TIME_SERIES will be used.
        feature_mask_fn_selector (Optional[FunctionSelector]): Selector of feature masks for the modality's explainers. If None selected, FEATURE_MASK_FUNCTIONS_FOR_TIME_SERIES will be used.
        pooling_fn_selector (Optional[FunctionSelector]): Selector of pooling methods for the modality's explainers. If None selected, POOLING_FUNCTIONS_FOR_TIME_SERIES will be used.
        normalization_fn_selector (Optional[FunctionSelector]): Selector of normalization methods for the modality's explainers. If None selected, NORMALIZATION_FUNCTIONS_FOR_TIME_SERIES will be used.
    """
    def __init__(
        self,
        channel_dim: int = -1,
        baseline_fn_selector: Optional[FunctionSelector] = None,
        feature_mask_fn_selector: Optional[FunctionSelector] = None,
        pooling_fn_selector: Optional[FunctionSelector] = None,
        normalization_fn_selector: Optional[FunctionSelector] = None,
    ):
        super(TimeSeriesModality, self).__init__(
            channel_dim,
            baseline_fn_selector=baseline_fn_selector or FunctionSelector(
                data=BASELINE_FUNCTIONS_FOR_TIME_SERIES, # [zeros, mean]
                default_kwargs={'dim': channel_dim},
            ),
            feature_mask_fn_selector=feature_mask_fn_selector or FunctionSelector(
                data=FEATURE_MASK_FUNCTIONS_FOR_TIME_SERIES,
            ),
            pooling_fn_selector=pooling_fn_selector or FunctionSelector(
                data=POOLING_FUNCTIONS_FOR_TIME_SERIES, # [identity]
                default_kwargs={'channel_dim': channel_dim},
            ),
            normalization_fn_selector=normalization_fn_selector or FunctionSelector(
                data=NORMALIZATION_FUNCTIONS_FOR_TIME_SERIES, # [identity]
            ),
        )

    def get_default_baseline_fn(self) -> BaselineFunction:
        """
        Defines default baseline function for the modality's explainers.

        Returns:
            BaselineFunction: Zeros baseline function.
        """
        return self.baseline_fn_selector.select('zeros')

    def get_default_feature_mask_fn(self) -> FeatureMaskFunction:
        """
        Defines default feature mask function for the modality's explainers.

        Returns:
            FeatureMaskFunction: No Mask baseline function.
        """
        return self.feature_mask_fn_selector.select('no_mask_2d')

    def get_default_postprocessors(self) -> List[PostProcessor]:
        """
        Defines default post-processors list for the modality's explainers.

        Returns:
            List[PostProcessor]: Identity PostProcessors.
        """
        return [
            PostProcessor(
                pooling_fn=self.pooling_fn_selector.select(pm),
                normalization_fn=self.normalization_fn_selector.select(nm),
            ) for pm in self.pooling_fn_selector.choices
            for nm in self.normalization_fn_selector.choices
        ]



from typing import Literal

import optuna


AVAILABLE_SAMPLERS = {
    'grid': optuna.samplers.GridSampler,
    'random': optuna.samplers.RandomSampler,
    'tpe': optuna.samplers.TPESampler,
}

DEFAULT_N_TRIALS = {
    'grid': None,
    'random': 50,
    'tpe': 50,
}

def load_sampler(sampler: Literal['grid', 'random', 'tpe']='tpe', **kwargs):
    return AVAILABLE_SAMPLERS[sampler](**kwargs)

def get_default_n_trials(sampler):
    return DEFAULT_N_TRIALS[sampler]

def nest_params(flattened_params):
    nested = {}
    for k, v in flattened_params.items():
        ref = nested
        splits = k.split('.')
        while splits:
            s = splits.pop(0)
            if s not in ref:
                _v = {} if len(splits) > 0 else v
                ref[s] = _v
            ref = ref[s]
    return nested



import warnings
from typing import Optional, Union, Callable

import torch

from pnpxai.core._types import Model
from pnpxai.explainers.base import Explainer
from pnpxai.evaluator.metrics.base import Metric


class Sensitivity(Metric):
    """
    Computes the complexity of attributions.
    
    Given `attributions`, calculates a fractional contribution distribution `prob_mass`,
    ``prob_mass[i] = hist[i] / sum(hist)``. where ``hist[i] = histogram(attributions[i])``.

    The complexity is defined by the entropy,
    ``evaluation = -sum(hist * ln(hist))``
    
    
    Args:
        model (Model): The model used for evaluation
        explainer (Optional[Union[Explainer, Callable]]): The explainer used for evaluation. It can be an instance of ``Explainer`` or any callable returning attributions from inputs and targets.
        n_iter (Optional[int]): The number of iterations for perturbation.
        epsilon (Optional[float]): The magnitude of random uniform noise.
    """
    def __init__(
        self,
        model: Model,
        explainer: Optional[Union[Explainer, Callable]] = None,
        n_iter: Optional[int] = 8,
        epsilon: Optional[float] = 0.2,
    ):
        super().__init__(model, explainer)
        self.n_iter = n_iter
        self.epsilon = epsilon

    def evaluate(
        self,
        inputs: torch.Tensor,
        targets: torch.Tensor,
        attributions: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Args:
            inputs (torch.Tensor): The input data.
            targets (torch.Tensor): The target labels for the inputs.
            attributions (Optional[torch.Tensor]): The attributions of the inputs.

        Returns:
            torch.Tensor: The result of the metric evaluation.
        """
        if self.explainer is None:
            warnings.warn('[Sensitivity] explainer is not provided. Please set explainer before evaluate.')
        explain_func = self.explainer.attribute if isinstance(self.explainer, Explainer) else self.explainer
        if attributions is None:
            attributions = explain_func(inputs, targets)
            # attributions = self.explainer.attribute(inputs, targets)
        attributions = attributions.to(self.device)
        evaluations = []
        for inp, target, attr in zip(inputs, targets, attributions):
            # Add random uniform noise which ranges [-epsilon, epsilon]
            perturbed = torch.stack([inp]*self.n_iter)
            noise = (
                torch.rand_like(perturbed).to(self.device) * self.epsilon * 2 \
                - self.epsilon
            )
            perturbed += noise
            # Get perturbed attribution results
            perturbed_attr = explain_func(
                perturbed.to(self.device),
                target.repeat(self.n_iter),
            )
            # Get maximum of the difference between the perturbed attribution and the original attribution
            attr_norm = torch.linalg.norm(attr).to(self.device)
            attr_diff = attr.to(self.device) - perturbed_attr.to(self.device)
            sens = max([torch.linalg.norm(diff)/attr_norm for diff in attr_diff])
            evaluations.append(sens)
        return torch.stack(evaluations).to(self.device)




import abc
import sys
import warnings
from typing import Optional, Union, Callable, Tuple

import copy
import torch
from torch import nn

from pnpxai.core._types import ExplanationType
from pnpxai.explainers import GradCam
from pnpxai.explainers.base import Explainer
from pnpxai.explainers.utils.postprocess import PostProcessor

# Ensure compatibility with Python 2/3
ABC = abc.ABC if sys.version_info >= (3, 4) else abc.ABCMeta(str("ABC"), (), {})

NON_DISPLAYED_ATTRS = [
    "model",
    "explainer",
    "device",
    "prob_fn",
    "pred_fn",
]


class Metric(ABC):
    """
    An abstract base class representing a metric used to evaluate explanations generated
    by an explainer for a given model. The metric is typically used to assess the quality or
    effectiveness of attribution-based explanations.

    Parameters:
        model (nn.Module): The model to be explained.
        explainer (Optional[Explainer], optional):
            The explainer used to generate explanations for the model. Defaults to None.
        **kwargs:
            Additional keyword arguments.
    """

    SUPPORTED_EXPLANATION_TYPE: ExplanationType = "attribution"

    def __init__(
        self, model: nn.Module, explainer: Optional[Explainer] = None, **kwargs
    ):
        self.model = model.eval()  # Set the model to evaluation mode
        self.explainer = explainer
        self.device = next(
            model.parameters()
        ).device  # Determine the device used by the model

    def __repr__(self):
        """
        Provides a string representation of the Metric object, displaying its attributes.

        Returns:
            str: A string representation of the Metric object.
        """
        displayed_attrs = ", ".join(
            [
                f"{k}={v}"
                for k, v in self.__dict__.items()
                if k not in NON_DISPLAYED_ATTRS and v is not None
            ]
        )
        return f"{self.__class__.__name__}({displayed_attrs})"

    def copy(self):
        """
        Creates a shallow copy of the Metric object.

        Returns:
            Metric: A copy of the Metric object.
        """
        return copy.copy(self)

    def set_explainer(self, explainer: Explainer):
        """
        Sets the explainer for the metric, ensuring it is associated with the same model.
        """
        assert self.model is explainer.model, "Must have same model of metric."
        clone = self.copy()
        clone.explainer = explainer
        return clone

    def set_kwargs(self, **kwargs):
        """
        Sets additional attributes for the Metric object using keyword arguments.
        """
        for k, v in kwargs.items():
            setattr(self, k, v)
        return self

    def evaluate(
        self,
        inputs: torch.Tensor,
        targets: torch.Tensor,
        attributions: Optional[torch.Tensor] = None,
        **kwargs,
    ) -> torch.Tensor:
        """
        Abstract method to evaluate the metric based on inputs, targets, and attributions.

        Parameters:
            inputs (Union[torch.Tensor, None]):
                The input data for the model.
            targets (Union[torch.Tensor, None]):
                The target labels for the input data.
            attributions (Union[torch.Tensor, None]):
                The attributions generated by the explainer.
            **kwargs:
                Additional keyword arguments.

        Returns:
            torch.Tensor: The result of the metric evaluation.
        """
        raise NotImplementedError

    def _get_attributions(
        self,
        inputs: torch.Tensor,
        targets: torch.Tensor,
        attributions: Optional[torch.Tensor] = None,
    ):
        if attributions is None:
            if self.explainer is None:
                raise Exception(
                    "[Sensitivity] Explainer must be set, if the attributions are not provided"
                )
            attributions = self.explainer.attribute(inputs, targets)

        return attributions



from typing import Optional, Sequence, Union, List, Tuple, Callable, Any, Dict, Literal

from collections import defaultdict
import threading

import torch
import torchvision.transforms.functional as TF
from torch import Tensor, device
from torch.nn import Module
from captum._utils.common import _run_forward, _sort_key_list, _reduce_list
from captum._utils.gradient import (
    compute_layer_gradients_and_eval,
    compute_gradients,
    apply_gradient_requirements,
    _extract_device_ids,
)
from zennit.attribution import Gradient as ZennitGradient
from zennit.core import Composite, Hook
from zennit.composites import LayerMapComposite
from zennit.types import Convolution, BatchNorm

from pnpxai.utils import format_into_tuple
from pnpxai.core._types import TensorOrTupleOfTensors, Tensor


class Gradient(ZennitGradient):
    def __init__(
        self,
        model: Module,
        composite: Optional[Composite]=None,
        attr_output=None,
        create_graph=False,
        retain_graph=None,
    ) -> None:
        super().__init__(model, composite, attr_output, create_graph, retain_graph)

    def grad(self, forward_args, targets, additional_forward_args=None):
        self._process_forward_args_before_grad(forward_args)
        grads = compute_gradients(
            self.model,
            forward_args,
            targets,
            additional_forward_args=additional_forward_args,
        )
        if len(grads) == 1:
            return grads[0]
        grads = tuple(g[0] for g in grads)
        return grads

    def forward(self, forward_args, targets, additional_forward_args=None):
        forward_args = self._process_forward_args_before_forward(forward_args)
        return self.grad(forward_args, targets, additional_forward_args)

    def _process_forward_args_before_forward(self, forward_args):
        if isinstance(forward_args, Sequence):
            return tuple(
                forward_arg.view_as(forward_arg) for forward_arg in forward_args
            )
        else:
            return forward_args.view_as(forward_args)

    def _process_forward_args_before_grad(self, forward_args):
        if isinstance(forward_args, Sequence):
            for forward_arg in forward_args:
                if forward_arg.is_floating_point() or forward_arg.is_complex():
                    forward_arg.requires_grad_()
        else:
            if forward_args.is_floating_point() or forward_args.is_complex():
                forward_args.requires_grad_()
        return forward_args


class LayerGradient(Gradient):
    def __init__(
        self,
        model: Module,
        layer: Union[str, Module, Sequence[Union[str, Module]]],
        composite: Optional[Composite]=None,
        attr_output=None,
        create_graph=False,
        retain_graph=None,
    ) -> None:
        super().__init__(model, composite, attr_output, create_graph, retain_graph)
        self.layer = layer

    def grad(self, forward_args, targets, additional_forward_args=None):
        grads, _ = compute_layer_gradients_and_eval(
            self.model,
            self.layer,
            forward_args,
            targets,
            additional_forward_args=additional_forward_args
        )
        if len(grads) == 1:
            return grads[0]
        grads = tuple(g[0] for g in grads)
        return grads


class SmoothGradient(Gradient):
    def __init__(
        self,
        model: Module,
        noise_level: Union[float, List[float]]=.1,
        n_iter: int=20,
        composite: Optional[Composite]=None,
        attr_output=None,
        create_graph=None,
        retain_graph=None,
    ) -> None:
        super().__init__(model, composite, attr_output, create_graph, retain_graph)
        self.noise_level = noise_level
        self.n_iter = n_iter

    def forward(
        self,
        inputs: TensorOrTupleOfTensors,
        targets: Tensor,
        additional_forward_args: Optional[TensorOrTupleOfTensors]=None,
        return_squared: bool=False,
    ):
        dims = tuple(range(1, inputs.ndim))
        std = self.noise_level * (inputs.amax(dims, keepdim=True) - inputs.amin(dims, keepdim=True))

        result = torch.zeros_like(inputs)
        result_sq = torch.zeros_like(inputs)
        for n in range(self.n_iter):
            if n == self.n_iter - 1:
                epsilon = torch.zeros_like(inputs)
            else:
                epsilon = torch.randn_like(inputs) * std
            grad = self.grad(
                inputs + epsilon,
                targets,
                additional_forward_args,
            )
            result += grad / self.n_iter
            if return_squared:
                result_sq += grad.pow(2) / self.n_iter
        if return_squared:
            return result, result_sq
        return result


class LayerSmoothGradient(LayerGradient):
    def __init__(
        self,
        model: Module,
        layer: Union[str, Module, List[Union[str, Module]]],
        noise_level: float=.1,
        n_iter: int=20,
        composite: Optional[Composite]=None,
        attr_output=None,
        create_graph=None,
        retain_graph=None,
    ) -> None:
        super().__init__(model, layer, composite, attr_output, create_graph, retain_graph)
        self.noise_level = noise_level
        self.n_iter = n_iter

    def grad(self, forward_args, targets, noise_fn, additional_forward_args=None):
        forward_args = self._process_forward_args_before_grad(forward_args)
        grads, _ = compute_layer_gradients_and_eval_with_noise(
            forward_fn=self.model,
            inputs=forward_args,
            layer=self.layer,
            noise_level=self.noise_level,
            noise_fn=noise_fn,
            target_ind=targets,
            additional_forward_args=additional_forward_args,
        )
        if len(grads) == 1:
            return grads
        grads = tuple(g[0] for g in grads)
        return grads

    def forward(
        self,
        forward_args: Union[Tensor, Tuple[Tensor]],
        targets: Tensor,
        additional_forward_args: Optional[Union[Tensor,Tuple[Tensor]]]=None,
        return_squared: bool=False,
    ) -> Union[Tensor, Tuple[Tensor]]:
        forward_args = self._process_forward_args_before_forward(forward_args)

        # forward to the layer
        forward_args = format_into_tuple(forward_args)
        for i in range(self.n_iter):
            noise_fn = torch.zeros_like if i == self.n_iter - 1 else torch.randn_like
            grads = self.grad(
                forward_args,
                targets,
                noise_fn,
                additional_forward_args=additional_forward_args,
            )
            if i == 0:
                results = tuple(grad/self.n_iter for grad in grads)
                if return_squared:
                    results_sq = tuple(grad.pow(2)/self.n_iter for grad in grads)
                continue
            results = tuple(
                result+grad/self.n_iter
                for result, grad in zip(results, grads)
            )
            if return_squared:
                results_sq = tuple(
                    result_sq+grad.pow(2)/self.n_iter
                    for result_sq, grad in zip(results, grads)
                )
        if return_squared:
            return results, results_sq
        return results


def _forward_layer_distributed_eval_with_noise(
    forward_fn: Callable,
    inputs: Any,
    layer: List[Union[str, Module]],
    noise_level: float,
    noise_fn: Callable=torch.randn_like,
    target_ind: Tensor = None,
    additional_forward_args: Any = None,
    attribute_to_layer_input: bool = False,
    forward_hook_with_return: bool = True,
    require_layer_grads: bool = True,
) -> Union[
    Tuple[Dict[Module, Dict[device, Tuple[Tensor, ...]]], Tensor],
    Dict[Module, Dict[device, Tuple[Tensor, ...]]],
]:
    r"""
    A helper function that allows to set a hook on model's `layer`, run the forward
    pass and returns intermediate layer results, stored in a dictionary,
    and optionally also the output of the forward function. The keys in the
    dictionary are the device ids and the values are corresponding intermediate layer
    results, either the inputs or the outputs of the layer depending on whether we set
    `attribute_to_layer_input` to True or False.
    This is especially useful when we execute forward pass in a distributed setting,
    using `DataParallel`s for example.
    """
    saved_layer: Dict[Module, Dict[device, Tuple[Tensor, ...]]] = defaultdict(dict)
    lock = threading.Lock()
    all_layers: List[Module] = [layer] if isinstance(layer, Module) else layer

    # Set a forward hook on specified module and run forward pass to
    # get layer output tensor(s).
    # For DataParallel models, each partition adds entry to dictionary
    # with key as device and value as corresponding Tensor.
    def hook_wrapper(original_module):
        def forward_hook(module, inp, out=None):
            eval_tsrs = inp if attribute_to_layer_input else out
            is_eval_tuple = isinstance(eval_tsrs, tuple)

            if not is_eval_tuple:
                eval_tsrs = (eval_tsrs,)

            # add noise
            dims = tuple(tuple(range(1, eval_tsr.ndim)) for eval_tsr in eval_tsrs)
            stds = tuple(
                noise_level * (eval_tsr.amax(dim=dim, keepdim=True) - eval_tsr.amin(dim=dim, keepdim=True))
                for eval_tsr, dim in zip(eval_tsrs, dims)
            )
            eval_tsrs = tuple(
                eval_tsr + noise_fn(eval_tsr) * std
                for eval_tsr, std in zip(eval_tsrs, stds)
            )

            if require_layer_grads:
                apply_gradient_requirements(eval_tsrs, warn=False)
            with lock:
                nonlocal saved_layer
                # Note that cloning behaviour of `eval_tsr` is different
                # when `forward_hook_with_return` is set to True. This is because
                # otherwise `backward()` on the last output layer won't execute.
                if forward_hook_with_return:
                    saved_layer[original_module][eval_tsrs[0].device] = eval_tsrs
                    eval_tsrs_to_return = tuple(
                        eval_tsr.clone() for eval_tsr in eval_tsrs
                    )
                    if not is_eval_tuple:
                        eval_tsrs_to_return = eval_tsrs_to_return[0]
                    return eval_tsrs_to_return
                else:
                    saved_layer[original_module][eval_tsrs[0].device] = tuple(
                        eval_tsr.clone() for eval_tsr in eval_tsrs
                    )

        return forward_hook

    all_hooks = []
    try:
        for single_layer in all_layers:
            if attribute_to_layer_input:
                all_hooks.append(
                    single_layer.register_forward_pre_hook(hook_wrapper(single_layer))
                )
            else:
                all_hooks.append(
                    single_layer.register_forward_hook(hook_wrapper(single_layer))
                )
        output = _run_forward(
            forward_fn,
            inputs,
            target=target_ind,
            additional_forward_args=additional_forward_args,
        )
    finally:
        for hook in all_hooks:
            hook.remove()

    if len(saved_layer) == 0:
        raise AssertionError("Forward hook did not obtain any outputs for given layer")

    if forward_hook_with_return:
        return saved_layer, output
    return saved_layer



def compute_layer_gradients_and_eval_with_noise(
    forward_fn: Callable,
    layer: List[Union[str, Module]],
    inputs: Union[Tensor, Tuple[Tensor, ...]],
    noise_level: float,
    noise_fn: Callable=torch.randn_like,
    target_ind: Optional[Tensor] = None,
    additional_forward_args: Any = None,
    gradient_neuron_selector: Union[
        None, int, Tuple[Union[int, slice], ...], Callable
    ] = None,
    device_ids: Union[None, List[int]] = None,
    attribute_to_layer_input: bool = False,
    output_fn: Union[None, Callable] = None,
    grad_kwargs: Optional[Dict[str, Any]] = None,
) -> Union[
    Tuple[Tuple[Tensor, ...], Tuple[Tensor, ...]],
    Tuple[Tuple[Tensor, ...], Tuple[Tensor, ...], Tuple[Tensor, ...]],
    Tuple[List[Tuple[Tensor, ...]], List[Tuple[Tensor, ...]]],
]:
    r"""
    Computes gradients of the output with respect to a given layer as well
    as the output evaluation of the layer for an arbitrary forward function
    and given input.

    For data parallel models, hooks are executed once per device ,so we
    need to internally combine the separated tensors from devices by
    concatenating based on device_ids. Any necessary gradients must be taken
    with respect to each independent batched tensor, so the gradients are
    computed and combined appropriately.

    More information regarding the behavior of forward hooks with DataParallel
    models can be found in the PyTorch data parallel documentation. We maintain
    the separate inputs in a dictionary protected by a lock, analogous to the
    gather implementation for the core PyTorch DataParallel implementation.

    NOTE: To properly handle inplace operations, a clone of the layer output
    is stored. This structure inhibits execution of a backward hook on the last
    module for the layer output when computing the gradient with respect to
    the input, since we store an intermediate clone, as
    opposed to the true module output. If backward module hooks are necessary
    for the final module when computing input gradients, utilize
    _forward_layer_eval_with_neuron_grads instead.

    Args:

        forward_fn: forward function. This can be for example model's
                    forward function.
        layer:      Layer for which gradients / output will be evaluated.
        inputs:     Input at which gradients are evaluated,
                    will be passed to forward_fn.
        target_ind: Index of the target class for which gradients
                    must be computed (classification only).
        output_fn:  An optional function that is applied to the layer inputs or
                    outputs depending whether the `attribute_to_layer_input` is
                    set to `True` or `False`
        args:       Additional input arguments that forward function requires.
                    It takes an empty tuple (no additional arguments) if no
                    additional arguments are required
        grad_kwargs: Additional keyword arguments for torch.autograd.grad


    Returns:
        tuple[**gradients**, **evals**]:
        - **gradients**:
            Gradients of output with respect to target layer output.
        - **evals**:
            Target layer output for given input.
    """
    with torch.autograd.set_grad_enabled(True):
        # saved_layer is a dictionary mapping device to a tuple of
        # layer evaluations on that device.
        saved_layer, output = _forward_layer_distributed_eval_with_noise(
            forward_fn,
            inputs,
            layer,
            noise_level,
            noise_fn,
            target_ind=target_ind,
            additional_forward_args=additional_forward_args,
            attribute_to_layer_input=attribute_to_layer_input,
            forward_hook_with_return=True,
            require_layer_grads=True,
        )
        assert output[0].numel() == 1, (
            "Target not provided when necessary, cannot"
            " take gradient with respect to multiple outputs."
        )

        device_ids = _extract_device_ids(forward_fn, saved_layer, device_ids)

        # Identifies correct device ordering based on device ids.
        # key_list is a list of devices in appropriate ordering for concatenation.
        # If only one key exists (standard model), key list simply has one element.
        key_list = _sort_key_list(
            list(next(iter(saved_layer.values())).keys()), device_ids
        )
        all_outputs: Union[Tuple[Tensor, ...], List[Tuple[Tensor, ...]]]
        if isinstance(layer, Module):
            all_outputs = _reduce_list(
                [
                    (
                        saved_layer[layer][device_id]
                        if output_fn is None
                        else output_fn(saved_layer[layer][device_id])
                    )
                    for device_id in key_list
                ]
            )
        else:
            all_outputs = [
                _reduce_list(
                    [
                        (
                            saved_layer[single_layer][device_id]
                            if output_fn is None
                            else output_fn(saved_layer[single_layer][device_id])
                        )
                        for device_id in key_list
                    ]
                )
                for single_layer in layer
            ]
        all_layers: List[Module] = [layer] if isinstance(layer, Module) else layer
        grad_inputs = tuple(
            layer_tensor
            for single_layer in all_layers
            for device_id in key_list
            for layer_tensor in saved_layer[single_layer][device_id]
        )
        saved_grads = torch.autograd.grad(
            outputs=torch.unbind(output),
            inputs=grad_inputs,
            **grad_kwargs or {},
        )

        offset = 0
        all_grads: List[Tuple[Tensor, ...]] = []
        for single_layer in all_layers:
            num_tensors = len(next(iter(saved_layer[single_layer].values())))
            curr_saved_grads = [
                saved_grads[i : i + num_tensors]
                for i in range(
                    offset, offset + len(key_list) * num_tensors, num_tensors
                )
            ]
            offset += len(key_list) * num_tensors
            if output_fn is not None:
                curr_saved_grads = [
                    output_fn(curr_saved_grad) for curr_saved_grad in curr_saved_grads
                ]

            all_grads.append(_reduce_list(curr_saved_grads))

        layer_grads: Union[Tuple[Tensor, ...], List[Tuple[Tensor, ...]]]
        layer_grads = all_grads
        if isinstance(layer, Module):
            layer_grads = all_grads[0]

        if gradient_neuron_selector is not None:
            assert isinstance(
                layer, Module
            ), "Cannot compute neuron gradients for multiple layers simultaneously!"
            inp_grads = _neuron_gradients(
                inputs, saved_layer[layer], key_list, gradient_neuron_selector
            )
            return (
                cast(Tuple[Tensor, ...], layer_grads),
                cast(Tuple[Tensor, ...], all_outputs),
                inp_grads,
            )
    return layer_grads, all_outputs  # type: ignore


def _get_bias_data(module):
    # Borrowed from official paper impl:
    # https://github.com/idiap/fullgrad-saliency/blob/master/saliency/tensor_extractor.py#L47
    if isinstance(module, BatchNorm):
        bias = -(module.running_mean*module.weight
            /(module.running_var+module.eps).sqrt())+module.bias
        return bias.data
    elif module.bias is not None:
        return module.bias.data
    return None


class _SavingBias(Hook):
    def forward(self, module, input, output):
        self.stored_tensors['bias'] = _get_bias_data(module)


class FullGradient(ZennitGradient):
    def __init__(
        self,
        model: Module,
        pooling_method: Optional[Callable[[Tensor], Tensor]]=None,
        interpolate_mode: Optional[TF.InterpolationMode]=None,
    ):
        layer_map = [(tp, _SavingBias()) for tp in [Convolution, BatchNorm]]
        composite = LayerMapComposite(layer_map=layer_map)
        super().__init__(model, composite, None)
        self.pooling_method = pooling_method
        self.interpolate_mode = interpolate_mode

    def forward(self, input, attr_output_fn):
        input = input.view_as(input)
        _, gradient = self.grad(input, attr_output_fn)
        rels = [gradient * input] # gradient x input
        for hook in self.composite.hook_refs:
            if hook.stored_tensors['bias'] is not None:
                grad_x_bias = (
                    hook.stored_tensors['bias'][None, :, None, None]
                    * hook.stored_tensors['grad_output'][0]
                )
                grad_x_bias = TF.resize(
                    grad_x_bias,
                    input.shape[2:],
                    interpolation=self.interpolate_mode,
                )
                rels.append(grad_x_bias)
        pooled = [self.pooling_method(rel) for rel in rels]
        return sum(pooled)[:, None, :, :]



from typing import Optional

import torch
from torchvision import transforms
from scipy.stats import spearmanr

from pnpxai.core._types import Model
from pnpxai.explainers.base import Explainer
from pnpxai.evaluator.metrics.base import Metric


class MuFidelity(Metric):
    """
    Computes the MuFidelity metric for attributions.

    Given a `model` and `inputs`, mufidelity of `model` to an explainer at `inputs` is calculated by
    a correlation between difference of predictions and attributions of maked inputs,
    ``evaluation = corr(pred_diff, masked_attr)``.

    The masked inputs are generated by masking `subset_mask` to noised `inputs`,
    ``masked = perturbed * subset_mask + (1.0 - subset_mask) * baseline``

    Args:
        model (Model): The model to evaluate.
        explainer (Optional[Explainer]): The explainer to evaluate.
        n_perturb (int): Number of perturbations to generate.
        noise_scale (int): Scale factor for Gaussian random noise.
        batch_size (int): Batch size for model evaluation.
        grid_size (int): Size of the grid for creating subsets.
        baseline (Union[float, torch.Tensor]): Baseline value for masked subsets.
        mask_agg_dim (Optional[int]): Dimension to aggregate masks.
        **kwargs: Additional kwargs to compute metric in an evaluator. Not required for single usage.

    Reference:
        U. Bhatt, A. Weller, and J. M. F. Moura. Evaluating and aggregating feature-based model attributions. In Proceedings of the IJCAI (2020).
    """

    def __init__(
        self,
        model: Model,
        explainer: Optional[Explainer] = None,
        n_perturb: int = 150,
        noise_scale: float = 0.2,
        batch_size: int = 32,
        grid_size: int = 9,
        baseline: float = 0.,
        mask_agg_dim: Optional[int] = None,
    ):
        super().__init__(model, explainer)
        self.n_perturb = n_perturb
        self.noise_scale = noise_scale
        self.batch_size = batch_size
        self.grid_size = grid_size
        self.baseline = baseline
        self.mask_agg_dim = mask_agg_dim

    def evaluate(
        self,
        inputs: torch.Tensor,
        targets: torch.Tensor,
        attributions: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Args:
            inputs (torch.Tensor): The input data (N x C x H x W).
            targets (torch.Tensor): The target labels for the inputs (N x 1).
            attributions (torch.Tensor): The attributions of the inputs.

        Returns:
            torch.Tensor: The result of the metric evaluation.
        """
        attributions = self._get_attributions(inputs, targets, attributions).to(self.device)

        outputs = self.model(inputs)
        n_classes = outputs.shape[-1]
        predictions = (
            outputs.to(self.device) * torch.eye(n_classes).to(self.device)[targets.to(self.device)]
        ).sum(dim=-1).detach()

        # input, target, attr, pred
        evaluations = []
        zipped = zip(inputs, targets, attributions, predictions)
        for input, target, attr, pred in zipped:
            repeated = torch.stack([input]*self.n_perturb).to(self.device)
            # Add Gaussian random noise
            std, mean = torch.std_mean(repeated)
            noise = torch.randn_like(repeated).to(self.device) * std.to(self.device) + mean.to(self.device)
            perturbed = self.noise_scale * noise + repeated
            perturbed = torch.minimum(repeated, perturbed)
            perturbed = torch.maximum(repeated-1, perturbed)

            # prepare the random masks that will designate the modified subset (S in original equation)
            subset_size = int(self.grid_size ** 2 * self.noise_scale)
            subset_mask = torch.randn(
                (self.n_perturb, self.grid_size ** 2)).to(self.device)
            subset_mask = torch.argsort(subset_mask, dim=-1) > subset_size
            subset_mask = torch.reshape(subset_mask.type(
                torch.float32), (self.n_perturb, 1, self.grid_size, self.grid_size))
            subset_mask = transforms.Resize(
                perturbed.shape[-2:],
                transforms.InterpolationMode("nearest")
            ).forward(subset_mask)
            if self.mask_agg_dim is not None:
                subset_mask = subset_mask.mean(dim=self.mask_agg_dim)

            # Use the masks to set the selected subsets to baseline state
            masked = perturbed * subset_mask + \
                (1.0 - subset_mask) * self.baseline
            masked = masked.to(self.device)

            masked_output = _forward_batch(self.model, masked, self.batch_size)
            pred_diff = pred - masked_output[:, target]

            masked_attr = (attr * (1.0 - subset_mask))\
                .sum(dim=tuple(range(1, subset_mask.ndim)))

            corr, _ = spearmanr(
                pred_diff.cpu().detach().numpy(),
                masked_attr.cpu().detach().numpy(),
            )
            evaluations.append(corr)
        return torch.tensor(evaluations)


def _forward_batch(model, inputs, batch_size) -> torch.Tensor:
    training_mode = model.training
    model.eval()
    device = next(model.parameters()).device
    outputs = []
    indices = list(range(len(inputs)))
    while indices:
        curr_indices, indices = indices[:batch_size], indices[batch_size:]
        outputs.append(model(inputs[curr_indices].to(device)))
    model.training = training_mode
    return torch.cat(outputs).detach()




from typing import Callable, Tuple, Union, Optional, Dict, Any

import torch
from torch import Tensor
from torch.nn.modules import Module
from captum.attr import Lime as CaptumLime

from pnpxai.core.detector.types import Linear, Convolution, LSTM, RNN, Attention
from pnpxai.explainers.base import Explainer
from pnpxai.explainers.utils.baselines import BaselineMethodOrFunction, BaselineFunction
from pnpxai.explainers.utils.feature_masks import FeatureMaskMethodOrFunction, FeatureMaskFunction
from pnpxai.utils import format_into_tuple


class Lime(Explainer):
    """
    Lime explainer.

    Supported Modules: `Linear`, `Convolution`, `LSTM`, `RNN`, `Attention`

    Parameters:
        model (Module): The PyTorch model for which attribution is to be computed.
        n_samples (int): Number of samples
        baseline_fn (Union[BaselineMethodOrFunction, Tuple[BaselineMethodOrFunction]]): The baseline function, accepting the attribution input, and returning the baseline accordingly.
        feature_mask_fn (Union[FeatureMaskMethodOrFunction, Tuple[FeatureMaskMethodOrFunction]): The feature mask function, accepting the attribution input, and returning the feature mask accordingly.
        perturb_fn (Optional[Callable[[Tensor], Tensor]]): The perturbation function, accepting the attribution input, and returning the perturbed value.
        forward_arg_extractor: A function that extracts forward arguments from the input batch(s) where the attribution scores are assigned.
        additional_forward_arg_extractor: A secondary function that extract additional forward arguments from the input batch(s).
        **kwargs: Keyword arguments that are forwarded to the base implementation of the Explainer

    Reference:
        Marco Tulio Ribeiro, Sameer Singh, Carlos Guestrin. "Why Should I Trust You?": Explaining the Predictions of Any Classifier.
    """

    SUPPORTED_MODULES = [Linear, Convolution, LSTM, RNN, Attention]

    def __init__(
        self,
        model: Module,
        n_samples: int = 25,
        baseline_fn: Union[BaselineMethodOrFunction,
                           Tuple[BaselineMethodOrFunction]] = 'zeros',
        feature_mask_fn: Union[FeatureMaskMethodOrFunction,
                               Tuple[FeatureMaskMethodOrFunction]] = 'felzenszwalb',
        perturb_fn: Optional[Callable[[Tensor], Tensor]] = None,
        forward_arg_extractor: Optional[Callable[[
            Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]] = None,
        additional_forward_arg_extractor: Optional[Callable[[
            Tuple[Tensor]], Tuple[Tensor]]] = None,
    ) -> None:
        super().__init__(model, forward_arg_extractor, additional_forward_arg_extractor)
        self.baseline_fn = baseline_fn or torch.zeros_like
        self.feature_mask_fn = feature_mask_fn
        self.perturb_fn = perturb_fn
        self.n_samples = n_samples

    def attribute(
            self,
            inputs: Tensor,
            targets: Optional[Tensor] = None,
    ) -> Union[Tensor, Tuple[Tensor]]:
        """
        Computes attributions for the given inputs and targets.

        Args:
            inputs (torch.Tensor): The input data.
            targets (torch.Tensor): The target labels for the inputs.

        Returns:
            Union[torch.Tensor, Tuple[torch.Tensor]]: The result of the explanation.
        """
        forward_args, additional_forward_args = self._extract_forward_args(
            inputs)
        forward_args = format_into_tuple(forward_args)

        explainer = CaptumLime(self.model, perturb_func=self.perturb_fn)
        attrs = explainer.attribute(
            inputs=forward_args,
            target=targets,
            baselines=self._get_baselines(forward_args),
            feature_mask=self._get_feature_masks(forward_args),
            n_samples=self.n_samples,
            additional_forward_args=additional_forward_args,
        )
        if isinstance(attrs, tuple) and len(attrs) == 1:
            attrs = attrs[0]
        return attrs

    
    def get_tunables(self) -> Dict[str, Tuple[type, Dict]]:
        """
        Provides Tunable parameters for the optimizer

        Tunable parameters:
            `n_samples` (int): Value can be selected in the range of `range(10, 100, 10)`

            `baseline_fn` (callable): BaselineFunction selects suitable values in accordance with the modality

            `feature_mask_fn` (callable): FeatureMaskFunction selects suitable values in accordance with the modality
        """
        return {
            'n_samples': (int, {'low': 10, 'high': 100, 'step': 10}),
            'baseline_fn': (BaselineFunction, {}),
            'feature_mask_fn': (FeatureMaskFunction, {})
        }



from typing import Type, Callable, Sequence, Dict, Any, Optional, Union

from torch.nn import ModuleDict
from torch.nn.modules import Module
from zennit.canonizers import AttributeCanonizer

from transformers.models.visual_bert.modeling_visual_bert import VisualBertAttention

class ModuleConverter:
    """
    Converts an input module into multiple output modules based on specified configurations.

    Args:
        in_module (Module): The input module to be converted.
        out_modules (Dict[str, Dict]): A dictionary specifying the configurations for output modules.
        output_selector (Callable[[Dict], Any] | None, optional): A function to select output from converted modules.
    """

    def __init__(
        self,
        in_module: Module,
        out_modules: Dict[str, Dict],
        output_selector: Optional[Callable[[Dict], Any]]=None,
    ):
        self.in_module = in_module
        self.out_modules = out_modules
        self.output_selector = output_selector


    def convert_module(self):
        """
        Converts the input module into multiple output modules based on the specified configurations.

        Returns:
            ModuleDict: A dictionary containing the converted output modules.
        """
        module_dict = ModuleDict()
        for out_module_nm, cfg in self.out_modules.items():
            if cfg.get("module"):
                get_out_module = cfg["module"]
                out_module = get_out_module(self.in_module)
                module_dict.update({out_module_nm: out_module})
                continue
            # create out module instance
            out_module_type = cfg["out_module_type"]
            out_module_args = {
                out_module_arg_nm: get_out_module_arg(self.in_module)
                if isinstance(get_out_module_arg, Callable) else get_out_module_arg
                for out_module_arg_nm, get_out_module_arg in cfg["args"].items()
            }
            out_module = out_module_type(**out_module_args)

            # update output params
            in_params = self.in_module.state_dict()
            out_params = out_module.state_dict()
            for out_param_nm, get_output_param in cfg["params"].items():
                if isinstance(get_output_param, Callable):
                    out_params[out_param_nm] = get_output_param(in_params)
                elif isinstance(get_output_param, str):
                    input_param = in_params.get(get_output_param)
                    if input_param is not None:
                        out_params[out_param_nm] = input_param
            out_module.load_state_dict(out_params)
            module_dict.update({out_module_nm: out_module})
        if not self.in_module.training:
            module_dict.eval()
        return module_dict


    def convert_forward_args(
        self,
        out_module_name,
        in_args,
        in_kwargs,
        kept,
        out_module
    ):
        """
        Converts input arguments for the forward method of the specified output module.

        Args:
            out_module_name (str): The name of the output module.
            in_args: Input arguments for the forward method of the input module.
            in_kwargs: Input keyword arguments for the forward method of the input module.
            kept: The kept tensors or values.

        Returns:
            tuple: A tuple containing:
                - forward_args: The converted input arguments for the forward method of the specified output module.
                - forward_kwargs: The converted input keyword arguments for the forward method of the specified output module.
        """
        cfg = self.out_modules[out_module_name]["forward"]
        forward_args = cfg["args"](in_args, in_kwargs, kept, out_module)
        if cfg.get("kwargs"):
            forward_kwargs = {
                k: get_kwarg(in_args, in_kwargs, kept, out_module)
                if isinstance(get_kwarg, Callable) else get_kwarg
                for k, get_kwarg in cfg["kwargs"].items()
            }
        else:
            forward_kwargs = {}
        return forward_args, forward_kwargs
    
    def keep_args(self, out_module_name):
        """
        Determines whether to keep input arguments for the forward method of the specified output module.

        Args:
            out_module_name (str): The name of the output module.

        Returns:
            bool: True if input arguments should be kept, False otherwise.
        """
        return self.out_modules[out_module_name]["forward"].get("keep_args")

    def keep_kwargs(self, out_module_name):
        """
        Determines whether to keep input keyword arguments for the forward method of the specified output module.

        Args:
            out_module_name (str): The name of the output module.

        Returns:
            bool: True if input keyword arguments should be kept, False otherwise.
        """
        return self.out_modules[out_module_name]["forward"].get("keep_kwargs")
    
    def keep_outputs(self, out_module_name):
        """
        Determines whether to keep output tensors from the forward method of the specified output module.

        Args:
            out_module_name (str): The name of the output module.

        Returns:
            bool: True if output tensors should be kept, False otherwise.
        """
        return self.out_modules[out_module_name]["forward"].get("keep_outputs")

    def select_output(self, kept):
        """
        Selects the output from converted modules based on the specified output selector function.

        Args:
            kept: The kept tensors or values.

        Returns:
            Any: The selected output.
        """
        return self.output_selector(kept)
    

def _validate_and_warn(*args, cfg=None):
    if cfg.get("validations"):
        for validate in cfg["validations"]:
            validate(*args)
    if cfg.get("warnings"):
        for warn in cfg["warnings"]:

            warn(*args)


def _converted_forward(self, *in_args, **in_kwargs):
    """
    Forwards input to the converted modules.

    Args:
        *in_args: Input arguments for the forward method.
        **in_kwargs: Input keyword arguments for the forward method.

    Returns:
        Any: The output from the converted modules.
    """
    kept_args: Dict[str, Any] = {}
    kept_kwargs: Dict[str, Any] = {}
    kept_outputs: Dict[str, Any] = {}
    kept = {
        "args": kept_args,
        "kwargs": kept_kwargs,
        "outputs": kept_outputs,
    }
    for out_module_nm, out_module in self._converted_module.items():
        out_module_args, out_module_kwargs = self._converter.convert_forward_args(
            out_module_nm, in_args, in_kwargs, kept, self._converted_module,
        )
        out_module_outputs = out_module(*out_module_args, **out_module_kwargs)
        if self._converter.keep_args(out_module_nm):
            kept_args[out_module_nm] = out_module_args
        if self._converter.keep_kwargs(out_module_nm):
            kept_kwargs[out_module_nm] = out_module_kwargs
        if self._converter.keep_outputs(out_module_nm):
            kept_outputs[out_module_nm] = out_module_outputs
    
    if self._converter.output_selector is not None:
        outputs = self._converter.output_selector(kept)
        return outputs
    return out_module_outputs



# for typing
class ModuleConvertingCanonizer(AttributeCanonizer):
    """
    Canonizer for converting input module to multiple output modules.

    Attributes:
        in_module_type (Module): The input module type.
        config_selector (Callable): A function to select the configuration for output modules.
        out_module_configs (Dict): A dictionary specifying the configurations for output modules.
    """

    in_module_type: Module
    config_selector: Callable
    out_module_configs: Dict

    def __init__(self, attribute_map):
        super().__init__(attribute_map)
    

# factory
def module_converting_canonizer_factory(
        in_module_type: Type[Module],
        config_selector: Callable[[Module], Union[bool, Sequence[bool]]],
        out_module_configs: Dict[Union[bool, Sequence[bool]], Dict[str, Any]],
    ):
    """
    Factory function to create a ModuleConvertingCanonizer subclass.

    Args:
        in_module_type (Type[Module]): The input module type.
        config_selector (Callable[[Module], bool|Sequence[bool]]): A function to select the configuration for output modules.
        out_module_configs (Dict[bool|Sequence[bool], Dict[str, Any]]): A dictionary specifying the configurations for output modules.

    Returns:
        Type[ModuleConvertingCanonizer]: A subclass of ModuleConvertingCanonizer.
    """

    class ModuleConvertingCanonizerFromFactory(ModuleConvertingCanonizer):
        """
        Canonizer generated from the factory to convert input module to multiple output modules.
        Inherits from ModuleConvertingCanonizer.
        """
        def __init__(self):
            super().__init__(self._attribute_map)
        
        @classmethod
        def _attribute_map(cls, name, in_module):
            """
            Maps attributes for conversion.

            Args:
                name: The name of the input module.
                in_module: The input module to be converted.

            Returns:
                dict or None: Attributes for conversion.
            """
            if isinstance(in_module, cls.in_module_type):
                cfg_key = cls.config_selector(in_module)
                out_module_configs = cls.out_module_configs[cfg_key]
                converter = ModuleConverter(
                    in_module=in_module,
                    out_modules=out_module_configs["out_modules"],
                    output_selector=out_module_configs["output_selector"],
                )
                converted_module = converter.convert_module()
                attributes = {
                    "forward": cls.forward.__get__(in_module),
                    "_converter": converter,
                    "_converted_module": converted_module,
                }
                return attributes
            return None

        @staticmethod
        def forward(self, *in_args, **in_kwargs):
            return _converted_forward(self, *in_args, **in_kwargs)

    ModuleConvertingCanonizerFromFactory.in_module_type = in_module_type
    ModuleConvertingCanonizerFromFactory.config_selector = config_selector
    ModuleConvertingCanonizerFromFactory.out_module_configs = out_module_configs
    return ModuleConvertingCanonizerFromFactory



from dataclasses import dataclass

import optuna
from pnpxai.explainers.base import Explainer
from pnpxai.explainers.utils.postprocess import PostProcessor


@dataclass
class OptimizationOutput:
    """
    A dataclass used to store the results of an optimization process, including the chosen 
    explainer, postprocessor, and the Optuna study object that was used during the optimization.

    Attributes:
        explainer (Explainer):
            The explainer object selected during the optimization process. This object is 
            responsible for generating explanations or attributions for a model.
        postprocessor (PostProcessor):
            The postprocessor object selected during the optimization process. This object 
            is used to process the attributions generated by the explainer, often involving 
            normalization or other transformations.
        study (optuna.study.Study):
            The Optuna study object that was used to conduct the optimization. This object 
            contains information about the optimization trials and results.
    """
    
    explainer: Explainer
    postprocessor: PostProcessor
    study: optuna.study.Study



from typing import Dict

MESSAGES: Dict[str, str] = {
    'experiment.event.explainer': 'Explaining with {explainer}.',
    'experiment.event.explainer.metric': 'Evaluating {metric} of {explainer}.',
    'experiment.errors.evaluation': 'Warning: Evaluating {metric} of {explainer} produced an error: {error}.',
    'experiment.errors.explanation': 'Warning: Explaining {explainer} produced an error: {error}.',
    'experiment.errors.explainer_unsupported': 'Warning: {explainer} is not currently supported.',
    'elapsed': 'Computed {modality} in {elapsed} sec',
    'project.config.unsupported': 'Error: Config of type {config_type} is not supported'
}


def get_message(key: str, *args, **kwargs):
    pattern = MESSAGES.get(key, None)
    return pattern.format(*args, **kwargs) if pattern is not None else key



from torch import Tensor
from pnpxai.explainers.utils.base import UtilFunction


def sumpos(attrs: Tensor, channel_dim: int) -> Tensor:
    return attrs.sum(channel_dim).clamp(min=0)


def sumabs(attrs: Tensor, channel_dim: int) -> Tensor:
    return attrs.sum(channel_dim).abs()


def l1norm(attrs: Tensor, channel_dim: int) -> Tensor:
    return attrs.abs().sum(channel_dim)


def maxnorm(attrs: Tensor, channel_dim: int) -> Tensor:
    return attrs.abs().max(channel_dim)[0]


def l2norm(attrs: Tensor, channel_dim: int) -> Tensor:
    return attrs.pow(2).sum(channel_dim).sqrt()


def l2normsq(attrs: Tensor, channel_dim: int) -> Tensor:
    return attrs.pow(2).sum(channel_dim)


def possum(attrs: Tensor, channel_dim: int) -> Tensor:
    return attrs.clamp(min=0).sum(channel_dim)


def posmaxnorm(attrs: Tensor, channel_dim: int) -> Tensor:
    return attrs.clamp(min=0).max(channel_dim)[0]


def posl2norm(attrs: Tensor, channel_dim: int) -> Tensor:
    return attrs.clamp(min=0).pow(2).sum(channel_dim).sqrt()


def posl2normsq(attrs: Tensor, channel_dim: int) -> Tensor:
    return attrs.clamp(min=0).pow(2).sum(channel_dim)


def identity(attrs: Tensor, *args, **kwargs) -> Tensor:
    return attrs


class PoolingFunction(UtilFunction):
    """
    A base class for pooling functions that aggregate or summarize attribution data across 
    certain dimensions. Pooling functions are often used to reduce the dimensionality of 
    attributions and highlight important features.

    Parameters:
        channel_dim (int):
            The dimension of the input channels. This dimension is used by the pooling function 
            to perform aggregation operations correctly.

    Notes:
        - `PoolingFunction` is intended to be subclassed. Concrete pooling methods should 
          inherit from this class and implement the actual pooling logic.
        - The pooling operation should be compatible with the `channel_dim` provided during 
          initialization.
    """
    
    def __init__(self, channel_dim: int):
        super().__init__()
        self.channel_dim = channel_dim


class SumPos(PoolingFunction):
    def __init__(self, channel_dim):
        super().__init__(channel_dim)

    def __call__(self, attrs: Tensor):
        return sumpos(attrs, self.channel_dim)


class SumAbs(PoolingFunction):
    def __init__(self, channel_dim):
        super().__init__(channel_dim)

    def __call__(self, attrs: Tensor):
        return sumabs(attrs, self.channel_dim)


class L1Norm(PoolingFunction):
    def __init__(self, channel_dim):
        super().__init__(channel_dim)

    def __call__(self, attrs: Tensor):
        return l1norm(attrs, self.channel_dim)


class MaxNorm(PoolingFunction):
    def __init__(self, channel_dim):
        super().__init__(channel_dim)

    def __call__(self, attrs: Tensor):
        return maxnorm(attrs, self.channel_dim)


class L2Norm(PoolingFunction):
    def __init__(self, channel_dim):
        super().__init__(channel_dim)

    def __call__(self, attrs: Tensor):
        return l2norm(attrs, self.channel_dim)


class L2NormSquare(PoolingFunction):
    def __init__(self, channel_dim):
        super().__init__(channel_dim)

    def __call__(self, attrs: Tensor):
        return l2normsq(attrs, self.channel_dim)


class PosSum(PoolingFunction):
    def __init__(self, channel_dim):
        super().__init__(channel_dim)

    def __call__(self, attrs: Tensor):
        return possum(attrs, self.channel_dim)


class PosMaxNorm(PoolingFunction):
    def __init__(self, channel_dim):
        super().__init__(channel_dim)

    def __call__(self, attrs: Tensor):
        return posmaxnorm(attrs, self.channel_dim)


class PosL2Norm(PoolingFunction):
    def __init__(self, channel_dim):
        super().__init__(channel_dim)

    def __call__(self, attrs: Tensor):
        return posl2norm(attrs, self.channel_dim)


class PosL2NormSquare(PoolingFunction):
    def __init__(self, channel_dim):
        super().__init__(channel_dim)

    def __call__(self, attrs: Tensor):
        return posl2normsq(attrs, self.channel_dim)


class Identity(UtilFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, inputs: Tensor):
        return identity(inputs)


POOLING_FUNCTIONS_FOR_IMAGE = {
    'sumpos': SumPos,
    'sumabs': SumAbs,
    'l1norm': L1Norm,
    'maxnorm': MaxNorm,
    'l2norm': L2Norm,
    'l2normsq': L2NormSquare,
    'possum': PosSum,
    'posmaxnorm': PosMaxNorm,
    'posl2norm': PosL2Norm,
    'posl2normsq': PosL2NormSquare,
}

POOLING_FUNCTIONS_FOR_TEXT = POOLING_FUNCTIONS_FOR_IMAGE

POOLING_FUNCTIONS_FOR_TIME_SERIES = {'identity': Identity}

POOLING_FUNCTIONS = {
    **POOLING_FUNCTIONS_FOR_IMAGE,
    **POOLING_FUNCTIONS_FOR_TEXT,
    **POOLING_FUNCTIONS_FOR_TIME_SERIES,
}


def minmax(attrs: Tensor):
    agg_dims = tuple(range(1, attrs.ndim))
    attrs_min = attrs.amin(agg_dims, keepdim=True)
    attrs_max = attrs.amax(agg_dims, keepdim=True)
    return (attrs - attrs_min) / (attrs_max - attrs_min)


class NormalizationFunction(UtilFunction):
    """
    A base class for normalization functions that adjust or scale attribution data. Normalization 
    functions are typically used to bring the data into a specific range or format for better 
    interpretation or visualization.

    Notes:
        - `NormalizationFunction` is designed to be subclassed. Concrete normalization methods 
          should inherit from this class and implement the actual normalization logic.
        - Subclasses can override the `__init__` method to accept additional parameters required 
          for their specific normalization operations.
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class MinMax(NormalizationFunction):
    def __init__(self):
        super().__init__()

    def __call__(self, attrs: Tensor):
        return minmax(attrs)


NORMALIZATION_FUNCTIONS_FOR_IMAGE = {
    'minmax': MinMax,
    'identity': Identity,
}

NORMALIZATION_FUNCTIONS_FOR_TEXT = NORMALIZATION_FUNCTIONS_FOR_IMAGE

NORMALIZATION_FUNCTIONS_FOR_TIME_SERIES = {
    'identity': Identity,
}

NORMALIZATION_FUNCTIONS = {
    **NORMALIZATION_FUNCTIONS_FOR_IMAGE,
    **NORMALIZATION_FUNCTIONS_FOR_TEXT,
    **NORMALIZATION_FUNCTIONS_FOR_TIME_SERIES,
}


class PostProcessor(UtilFunction):
    """
    A class that applies a series of post-processing steps to the output of an attribution method.
    This includes pooling and normalization functions to refine and transform the attributions.

    Parameters:
        pooling_fn (PoolingFunction):
            A function used to perform pooling on the attributions. Pooling typically involves 
            aggregating or summarizing the attributions over certain dimensions.
        normalization_fn (NormalizationFunction):
            A function used to normalize the pooled attributions. Normalization typically involves 
            scaling or adjusting the attributions to a certain range or format.

    Methods:
        from_name(pooling_method: str, normalization_method: str, channel_dim: int) -> PostProcessor:
            Creates a `PostProcessor` instance using the specified method names for pooling and 
            normalization, and the channel dimension. This is a convenience method for instantiating 
            `PostProcessor` with predefined methods.

        __call__(attrs):
            Applies the pooling and normalization functions to the given attributions. The input 
            attributions are first pooled and then normalized.

        get_tunables() -> Dict[str, Tuple[Type, Dict[str, Any]]]:
            Returns a dictionary of tunable parameters for the `PostProcessor`. The dictionary 
            includes the functions for pooling and normalization along with their default parameters.
    """
    
    def __init__(
        self,
        pooling_fn: PoolingFunction,
        normalization_fn: NormalizationFunction,
    ):
        self.pooling_fn = pooling_fn
        self.normalization_fn = normalization_fn

    @classmethod
    def from_name(
        cls,
        pooling_method: str,
        normalization_method: str,
        channel_dim: int,
    ):
        return cls(
            pooling_fn=POOLING_FUNCTIONS[pooling_method](channel_dim),
            normalization_fn=NORMALIZATION_FUNCTIONS[normalization_method](),
        )

    def __call__(self, attrs):
        pooled = self.pooling_fn(attrs)
        normalized = self.normalization_fn(pooled)
        return normalized

    def get_tunables(self):
        return {
            'pooling_fn': (PoolingFunction, {}),
            'normalization_fn': (NormalizationFunction, {}),
        }



from typing import Sequence, Any, Union
import functools

import torch
from torch import nn
from captum.attr._utils.input_layer_wrapper import ModelInputWrapper
from skimage.segmentation import felzenszwalb

from pnpxai.core.detector import symbolic_trace
from pnpxai.core.detector.utils import get_target_module_of, find_nearest_user_of
from pnpxai.core.detector.types import Convolution, Pool


def find_cam_target_layer(model: nn.Module) -> nn.Module:
    traced_model = symbolic_trace(model)
    last_conv_node = next(
        node for node in reversed(traced_model.graph.nodes)
        if isinstance(get_target_module_of(node), Convolution)
    )
    if last_conv_node is None:
        return
    pool_user = find_nearest_user_of(last_conv_node, Pool)
    if pool_user is None:
        return get_target_module_of(last_conv_node)
    conv_module_nm = next(iter(pool_user.prev.meta.get("nn_module_stack")))
    target_module = model
    for t in conv_module_nm.split("."):
        target_module = getattr(target_module, t)
    return target_module


def default_feature_mask_fn_image(inputs: torch.Tensor, scale=250):
    feature_mask = [
        torch.tensor(felzenszwalb(input.permute(1,2,0).detach().cpu().numpy(), scale=scale))
        for input in inputs
    ]
    return torch.LongTensor(torch.stack(feature_mask)).to(inputs.device)


def default_feature_mask_fn_text(inputs): return None


def default_feature_mask_fn_image_text(images, text):
    fm_img = default_feature_mask_fn_image(images)
    bsz, text_len = text.size()
    fm_text = torch.arange(text_len).repeat(bsz).view(bsz, text_len)
    fm_text += fm_img.max().item() + 1
    return fm_img, fm_text


def get_default_feature_mask_fn(modality):
    if modality == 'image':
        return default_feature_mask_fn_image
    elif modality == 'text':
        return default_feature_mask_fn_text
    elif modality == ('image', 'text'):
        return default_feature_mask_fn_image_text
    else:
        raise NotImplementedError(f"default_feature_mask_fn for '{modality}' not supported.")


def default_baseline_fn_image(inputs: torch.Tensor):
    return torch.zeros_like(inputs)

def default_baseline_fn_text(inputs: torch.Tensor, mask_token_id: int=0):
    return torch.ones_like(inputs, dtype=torch.long) * mask_token_id

def default_baseline_fn_image_text(images: torch.Tensor, text: torch.Tensor, mask_token_id: int=0):
    return default_baseline_fn_image(images), default_baseline_fn_text(text, mask_token_id)

def get_default_baseline_fn(modality, mask_token_id=0):
    if modality == 'image':
        return default_baseline_fn_image
    elif modality == 'text':
        return default_baseline_fn_text
    elif modality == ('image', 'text'):
        return functools.partial(default_baseline_fn_image_text, mask_token_id=mask_token_id)
    else:
        raise NotImplementedError(f"default_baseline_fn for '{modality}' not supported.")

def captum_wrap_model_input(model):
    if isinstance(model, nn.DataParallel):
        return ModelInputWrapper(model.module)
    return ModelInputWrapper(model)



def _format_to_tuple(obj: Union[Any, Sequence[Any]]):
    if isinstance(obj, Sequence):
        return tuple(obj)
    return (obj,)


from typing import Callable, Optional, Tuple, List
import itertools

from torch.nn.modules import Module
from torch.utils.data import DataLoader

from pnpxai.core._types import Model, DataSource
from pnpxai.core.experiment.experiment import Experiment
from pnpxai.core.modality.modality import Modality, ImageModality, TextModality, TimeSeriesModality
from pnpxai.core.recommender import XaiRecommender
from pnpxai.explainers.types import TargetLayer
from pnpxai.evaluator.metrics import PIXEL_FLIPPING_METRICS
from pnpxai.evaluator.metrics import (
    MuFidelity,
    Sensitivity,
    Complexity,
    MoRF,
    LeRF,
    AbPC,
)
from pnpxai.utils import format_into_tuple


METRICS_BASELINE_FN_REQUIRED = PIXEL_FLIPPING_METRICS
METRICS_CHANNEL_DIM_REQUIRED = PIXEL_FLIPPING_METRICS
DEFAULT_METRICS_FOR_TEXT = [
    MoRF,
    LeRF,
    AbPC,
]
DEFAULT_METRICS = [
    MuFidelity,
    AbPC,
    Sensitivity,
    Complexity,
]


class AutoExplanation(Experiment):
    """
    An extension of Experiment class with automatic explainers and parameters recommendation.

    Parameters:
        model (Model): The machine learning model to be analyzed.
        data (DataSource): The data source used for the experiment.
        modality (Modality): An object to specify modality-specific workflow.
        input_extractor (Optional[Callable], optional): Custom function to extract input features.
        label_extractor (Optional[Callable], optional): Custom function to extract labels features.
        target_extractor (Optional[Callable], optional): Custom function to extract target features.
        input_visualizer (Optional[Callable], optional): Custom function for visualizing input features.
        target_labels (Optional[bool]): Whether to use target labels.

    Attributes:
        recommended (RecommenderOutput): A data object, containing recommended explainers.
    """

    def __init__(
        self,
        model: Model,
        data: DataSource,
        modality: Modality,
        input_extractor: Optional[Callable] = None,
        label_extractor: Optional[Callable] = None,
        target_extractor: Optional[Callable] = None,
        target_labels: bool = False
    ):
        self.recommended = XaiRecommender().recommend(modality=modality, model=model)
        self.modality = modality

        super().__init__(
            model=model,
            data=data,
            modality=modality,
            explainers=self._load_default_explainers(model),
            postprocessors=self._load_default_postprocessors(),
            metrics=self._load_default_metrics(model),
            input_extractor=input_extractor,
            label_extractor=label_extractor,
            target_extractor=target_extractor,
            target_labels=target_labels,
        )

    def _load_default_explainers(self, model):
        explainers = []
        for explainer_type in self.recommended.explainers:
            explainer = explainer_type(model=model)
            default_kwargs = self._generate_default_kwargs_for_explainer()
            for k, v in default_kwargs.items():
                if hasattr(explainer, k):
                    explainer = explainer.set_kwargs(**{k: v})
            explainers.append(explainer)
        return explainers

    def _load_default_metrics(self, model):
        empty_metrics = []  # empty means that explainer is not assigned yet
        for metric_type in DEFAULT_METRICS:
            metric = metric_type(model=model)
            default_kwargs = self._generate_default_kwargs_for_metric()
            for k, v in default_kwargs.items():
                if hasattr(metric, k):
                    metric = metric.set_kwargs(**{k: v})
            empty_metrics.append(metric)
        return empty_metrics

    def _load_default_postprocessors(self):
        modalities = format_into_tuple(self.modality)
        if len(modalities) == 1:
            return self.modality.get_default_postprocessors()
        return list(itertools.product(*tuple(
            modality.get_default_postprocessors()
            for modality in modalities
        )))

    def _generate_default_kwargs_for_explainer(self):
        return {
            'feature_mask_fn': self.modality.get_default_feature_mask_fn(),
            'baseline_fn': self.modality.get_default_baseline_fn(),
        }

    def _generate_default_kwargs_for_metric(self):
        return {
            'baseline_fn': self.modality.get_default_baseline_fn(),
            'channel_dim': self.modality.channel_dim,
        }


class AutoExplanationForImageClassification(AutoExplanation):
    """
    An extension of AutoExplanation class with modality set to the ImageModality.

    Parameters:
        model (Model): The machine learning model to be analyzed.
        data (DataSource): The data source used for the experiment.
        input_extractor (Optional[Callable]): Custom function to extract input features.
        label_extractor (Optional[Callable]): Custom function to extract labels features.
        target_extractor (Optional[Callable]): Custom function to extract target features.
        target_labels (Optional[bool]): Whether to use target labels.
        channel_dim (int): Channel dimension.

    Attributes:
        modality (ImageModality): An object to specify modality-specific workflow.
        recommended (RecommenderOutput): A data object, containing recommended explainers.
    """
    def __init__(
        self,
        model: Module,
        data: DataLoader,
        input_extractor: Optional[Callable] = None,
        label_extractor: Optional[Callable] = None,
        target_extractor: Optional[Callable] = None,
        target_labels: bool = False,
        channel_dim: int = 1,
    ):
        super().__init__(
            model=model,
            data=data,
            modality=ImageModality(channel_dim),
            input_extractor=input_extractor,
            label_extractor=label_extractor,
            target_extractor=target_extractor,
            target_labels=target_labels
        )


class AutoExplanationForTextClassification(AutoExplanation):
    """
    An extension of AutoExplanation class with modality set to the TextModality.

    Parameters:
        model (Model): The machine learning model to be analyzed.
        data (DataSource): The data source used for the experiment.
        layer (TargetLayer): A Module or its string representation to select a target layer for analysis.
        mask_token_id (int): A mask token id.
        input_extractor (Optional[Callable], optional): Custom function to extract input features.
        forward_arg_extractor (Optional[Callable]): Custom function to extract forward arguments.
        additional_forward_arg_extractor (Optional[Callable]): Custom function to extract additional forward arguments.
        label_extractor (Optional[Callable], optional): Custom function to extract labels features.
        target_extractor (Optional[Callable], optional): Custom function to extract target features.
        target_labels (Optional[bool]): Whether to use target labels.
        channel_dim (int): Channel dimension.

    Attributes:
        modality (ImageModality): An object to specify modality-specific workflow.
        recommended (RecommenderOutput): A data object, containing recommended explainers.
    """
    def __init__(
        self,
        model: Module,
        data: DataLoader,
        layer: TargetLayer,
        mask_token_id: int,
        input_extractor: Optional[Callable] = None,
        forward_arg_extractor: Optional[Callable] = None,
        additional_forward_arg_extractor: Optional[Callable] = None,
        label_extractor: Optional[Callable] = None,
        target_extractor: Optional[Callable] = None,
        target_labels: bool = False,
        channel_dim: int = -1,
    ):
        self.layer = layer
        self.mask_token_id = mask_token_id
        self.forward_arg_extractor = forward_arg_extractor
        self.additional_forward_arg_extractor = additional_forward_arg_extractor

        super().__init__(
            model=model,
            data=data,
            modality=TextModality(
                channel_dim=channel_dim, mask_token_id=mask_token_id),
            input_extractor=input_extractor,
            label_extractor=label_extractor,
            target_extractor=target_extractor,
            target_labels=target_labels
        )

    def _generate_default_kwargs_for_explainer(self):
        return {
            'layer': self.layer,
            'forward_arg_extractor': self.forward_arg_extractor,
            'additional_forward_arg_extractor': self.additional_forward_arg_extractor,
            'feature_mask_fn': self.modality.get_default_feature_mask_fn(),
            'baseline_fn': self.modality.get_default_baseline_fn(),
        }

    def _generate_default_kwargs_for_metric(self):
        return {
            'baseline_fn': self.modality.get_default_baseline_fn(),
            'channel_dim': self.modality.channel_dim,
        }

    def _load_default_metrics(self, model):
        empty_metrics = []  # empty means that explainer is not assigned yet
        for metric_type in DEFAULT_METRICS_FOR_TEXT:
            metric = metric_type(model=model)
            default_kwargs = self._generate_default_kwargs_for_metric()
            for k, v in default_kwargs.items():
                if hasattr(metric, k):
                    metric = metric.set_kwargs(**{k: v})
            empty_metrics.append(metric)
        return empty_metrics


class AutoExplanationForVisualQuestionAnswering(AutoExplanation):
    """
    An extension of AutoExplanation class with multiple modalities, namely ImageModality and TextModality.

    Parameters:
        model (Model): The machine learning model to be analyzed.
        data (DataSource): The data source used for the experiment.
        layer (TargetLayer): A Module or its string representation to select a target layer for analysis.
        mask_token_id (int): A mask token id.
        modality (Modality): An object to specify modality-specific workflow.
        input_extractor (Optional[Callable], optional): Custom function to extract input features.
        forward_arg_extractor (Optional[Callable]): Custom function to extract forward arguments.
        additional_forward_arg_extractor (Optional[Callable]): Custom function to extract additional forward arguments.
        label_extractor (Optional[Callable], optional): Custom function to extract labels features.
        target_extractor (Optional[Callable], optional): Custom function to extract target features.
        target_labels (Optional[bool]): Whether to use target labels.
        channel_dim (Tuple[int]): Channel dimension. Requires a tuple channel dimensions for image and text modalities.

    Attributes:
        modality (Tuple[ImageModality, TextModality]): A tuple of objects to specify modality-specific workflow.
        recommended (RecommenderOutput): A data object, containing recommended explainers.
    """
    def __init__(
        self,
        model: Module,
        data: DataLoader,
        layer: List[TargetLayer],
        mask_token_id: int,
        input_extractor: Optional[Callable] = None,
        forward_arg_extractor: Optional[Callable] = None,
        additional_forward_arg_extractor: Optional[Callable] = None,
        label_extractor: Optional[Callable] = None,
        target_extractor: Optional[Callable] = None,
        target_labels: bool = False,
        channel_dim: Tuple[int] = (1, -1),
    ):
        self.layer = layer
        self.mask_token_id = mask_token_id
        self.forward_arg_extractor = forward_arg_extractor
        self.additional_forward_arg_extractor = additional_forward_arg_extractor
        super().__init__(
            model=model,
            data=data,
            modality=(
                ImageModality(channel_dim=channel_dim[0]),
                TextModality(channel_dim=channel_dim[1], ),
            ),
            input_extractor=input_extractor,
            label_extractor=label_extractor,
            target_extractor=target_extractor,
            target_labels=target_labels
        )

    def _generate_default_kwargs_for_explainer(self):
        return {
            'layer': self.layer,
            'forward_arg_extractor': self.forward_arg_extractor,
            'additional_forward_arg_extractor': self.additional_forward_arg_extractor,
            'feature_mask_fn': tuple(
                modality.get_default_feature_mask_fn() for modality in self.modality
            ),
            'baseline_fn': tuple(
                modality.get_default_baseline_fn() for modality in self.modality
            ),
        }

    def _generate_default_kwargs_for_metric(self):
        return {
            'baseline_fn': tuple(
                modality.get_default_baseline_fn() for modality in self.modality
            ),
            'channel_dim': tuple(modality.channel_dim for modality in self.modality),
        }


class AutoExplanationForTSClassification(AutoExplanation):
    """
    An extension of AutoExplanation class with modality set to the TimeSeriesModality.

    Parameters:
        model (Model): The machine learning model to be analyzed.
        data (DataSource): The data source used for the experiment.
        input_extractor (Optional[Callable], optional): Custom function to extract input features.
        label_extractor (Optional[Callable], optional): Custom function to extract labels features.
        target_extractor (Optional[Callable], optional): Custom function to extract target features.
        target_labels (Optional[bool]): Whether to use target labels.
        sequence_dim (Tuple[int]): Sequence dimension.
        mask_agg_dim (Tuple[int]): A dimension for aggregating mask values. Usually, a channel dimension.

    Attributes:
        modality (TimeSeriesModality): An object to specify modality-specific workflow.
        recommended (RecommenderOutput): A data object, containing recommended explainers.
    """
    def __init__(
        self,
        model: Module,
        data: DataLoader,
        input_extractor: Optional[Callable] = None,
        label_extractor: Optional[Callable] = None,
        target_extractor: Optional[Callable] = None,
        target_labels: bool = False,
        sequence_dim: int = -1,
        mask_agg_dim: int = -2,
    ):
        self.mask_agg_dim = mask_agg_dim
        super().__init__(
            model=model,
            data=data,
            modality=TimeSeriesModality(sequence_dim),
            input_extractor=input_extractor,
            label_extractor=label_extractor,
            target_extractor=target_extractor,
            target_labels=target_labels
        )

    def _generate_default_kwargs_for_metric(self):
        return {
            'baseline_fn': self.modality.get_default_baseline_fn(),
            'feature_mask_fn': self.modality.get_default_feature_mask_fn(),
            'channel_dim': self.modality.channel_dim,
            'mask_agg_dim': self.mask_agg_dim,
        }



from zennit.core import RemovableHandleList, RemovableHandle, Hook, BasicHook

class HookWithKwargs(Hook):
    '''Base class for hooks to be used to compute layer-wise attributions.'''
    def __init__(self):
        super().__init__()
        self.stored_kwargs = None
    
    def pre_forward(self, module, input, kwargs):
        return super().pre_forward(module, input), kwargs

    def forward(self, module, input, kwargs, output):
        '''Forward hook to save module in-/outputs.'''
        self.stored_tensors['input'] = input
        self.stored_kwargs = kwargs

    def register(self, module):
        '''Register this instance by registering all hooks to the supplied module.'''
        return RemovableHandleList([
            RemovableHandle(self),
            module.register_forward_pre_hook(self.pre_forward, with_kwargs=True),
            module.register_forward_hook(self.post_forward),
            module.register_forward_hook(self.forward, with_kwargs=True),
        ])


# myeongjin hi
class BasicHookWithRelevanceModifier(BasicHook):
    def __init__(
    ) -> None:
        pass




from typing import Set, Tuple, Optional
from torch import fx, nn
from pnpxai.core._types import Model
from pnpxai.core.detector.utils import get_target_module_of
from pnpxai.core.detector.types import (
    ModuleType,
    Linear,
    Convolution,
    RNN,
    LSTM,
    Attention,
    Embedding,
)

DEFAULT_MODULE_TYPES_TO_DETECT = (
    Linear,
    Convolution,
    RNN,
    LSTM,
    Attention,
    Embedding,
)


class Tracer(fx.Tracer):
    # stop recursion when meeting timm's Attention
    def is_leaf_module(self, m: nn.Module, module_qualified_name: str) -> bool:
        return (
            (
                m.__module__.startswith("torch.nn")
                or m.__module__.startswith("torch.ao.nn")
                or (
                    m.__module__ == "timm.models.vision_transformer"
                    and m.__class__.__name__ == "Attention"
                )
            ) and not isinstance(m, nn.Sequential)
        )


def symbolic_trace(model: nn.Module) -> fx.GraphModule:
    tracer = Tracer()
    graph = tracer.trace(model)
    name = (
        model.__class__.__name__ if isinstance(
            model, nn.Module) else model.__name__
    )
    return fx.GraphModule(tracer.root, graph, name)


def _get_node_target_module(node):
    target_module = node.graph.owning_module
    for t in node.target.splits('.'):
        target_module = getattr(target_module, t)
    return target_module


def _format_node_target(node) -> str:
    if node.op == 'get_attr':
        return f'getattr(self, {node.target})'
    elif node.op == 'call_function':
        return str(node.target)
    elif node.op == 'call_module':
        target_module = node.graph.owning_module
        for t in node.target.split('.'):
            target_module = getattr(target_module, t)
        return str(target_module)
    return node.target


def extract_graph_data(graph_module: fx.GraphModule):
    nodes, edges = [], []
    data = {'nodes': nodes, 'edges': edges}
    for node in graph_module.graph.nodes:
        nodes.append({
            'name': node.name,
            'op': node.op,
            'target': _format_node_target(node),
        })
        edges += [{
            'source': node.name,
            'target': user.name,
        } for user in node.users]
    return data


def detect_model_architecture(
    model: Model,
    targets: Optional[Tuple[ModuleType]] = None,
) -> Set[ModuleType]:
    """
    A function detecting architecture for a given model.

    Args:
        model (Model): The machine learning model to be detected

    Returns:
        ModelArchitectureSummary: A summary of model architecture
    """
    targets = targets or DEFAULT_MODULE_TYPES_TO_DETECT
    detected = set()
    for nm, module in model.named_modules():
        module_type = next(
            (target for target in targets if isinstance(module, target)), None)
        if module_type is None:
            continue
        detected.add(module_type)
    return detected



from typing import List, Type, Dict, Set, Any, Sequence, Tuple, Union
from dataclasses import dataclass, asdict
from tabulate import tabulate

from pnpxai.core._types import Model
from pnpxai.core.modality.modality import Modality
from pnpxai.core.detector import detect_model_architecture
from pnpxai.core.detector.types import (
    ModuleType,
    Linear,
    Convolution
)
from pnpxai.explainers.base import Explainer
from pnpxai.explainers import (
    GradCam,
    GuidedGradCam,
    AVAILABLE_EXPLAINERS
)
from pnpxai.utils import format_into_tuple


CAM_BASED_EXPLAINERS = {GradCam, GuidedGradCam}


@dataclass
class RecommenderOutput:
    detected_architectures: Set[ModuleType]
    explainers: list

    def print_tabular(self):
        print(tabulate([
            [k, [v.__name__ for v in vs]]
            for k, vs in asdict(self).items()
        ]))


def _sort_by_name(vals):
    return list(sorted(vals, key=lambda x: x.__name__))


class RecommendationMap:
    def __init__(self, data: Dict[Any, Set[Any]], headers: Sequence[str]) -> None:
        self.data = data
        self.headers = headers

    def print_tabular(self):
        table = [self.headers]
        for k, vs in self.data.items():
            if not isinstance(k, (str, Tuple)):
                k = k.__name__
            table.append([k, ', '.join(sorted(v.__name__ for v in vs))])
        print(tabulate(table, headers="firstrow"))


class XaiRecommender:
    """
    Recommender class that suggests explainers and evaluation metrics based on the user's input. The recommender
    analyzes the model architecture and the selected modality to provide the most suitable explainers and detected architecture.        
    """

    def __init__(self):
        self.architecture_to_explainers_map = self._build_architecture_to_explainers_map()

    def _build_architecture_to_explainers_map(self):
        map_data = {}
        for explainer_type in AVAILABLE_EXPLAINERS:
            for arch in explainer_type.SUPPORTED_MODULES:
                if arch not in map_data:
                    map_data[arch] = set()
                map_data[arch].add(explainer_type)
        return RecommendationMap(map_data, ["architecture", "explainers"])

    def _filter_explainers(
        self,
        modality: Union[Modality, Tuple[Modality]],
        arch: Set[ModuleType],
    ) -> List[Type[Explainer]]:
        """
        Filters explainers based on the user's question, task, and model architecture.

        Args:
        - modaltiy (Union[Modality, Tuple[Modality]]): Modality of the input data (e.g., ImageModality, TextModality, TabularModality).
        - arch (Set[ModuleType]): Set of neural network module types (e.g., nn.Linear, nn.Conv2d).

        Returns:
        - List[Set[Type[Explainer]]]: List of compatible explainers based on the given inputs.
        """
        # question_to_method = QUESTION_TO_EXPLAINERS.get(question, set())
        explainers = []
        for mod in format_into_tuple(modality):
            modality_to_explainers = set(mod.EXPLAINERS)
            arch_to_explainers = set.union(*(
                self.architecture_to_explainers_map.data.get(
                    module_type, set())
                for module_type in arch
            ))
            explainers_mod = set.intersection(
                modality_to_explainers, arch_to_explainers)
            if arch.difference({Convolution, Linear}):
                explainers_mod = explainers_mod.difference(
                    CAM_BASED_EXPLAINERS)
            explainers.append(explainers_mod)
        explainers = set.intersection(*explainers)
        return list(explainers)

    def recommend(self, modality: Union[Modality, Tuple[Modality]], model: Model) -> RecommenderOutput:
        """
        Recommends explainers and evaluation metrics based on the user's input.

        Args:
        - modality (Union[Modality], Tuple[Modality]): Modality of the input data (e.g., ImageModality, TextModality, TabularModality).
        - model (Model): Neural network module, used for the architecture-based filtering.

        Returns:
        - RecommenderOutput: An object containing recommended explainers.
        """
        arch = detect_model_architecture(model)
        explainers = self._filter_explainers(modality, arch)
        # metrics = self._suggest_metrics(explainers)
        return RecommenderOutput(
            detected_architectures=arch,
            explainers=_sort_by_name(explainers),
            # metrics=_sort_by_name(metrics),
        )



from typing import Dict, Tuple
from torch import Tensor, nn
from captum.attr import LayerGradCam, LayerAttribution

from pnpxai.utils import format_into_tuple
from pnpxai.core.detector.types import Convolution
from pnpxai.explainers.base import Explainer
from pnpxai.explainers.utils import find_cam_target_layer
from pnpxai.explainers.errors import NoCamTargetLayerAndNotTraceableError


class GradCam(Explainer):
    """
    GradCAM explainer.

    Supported Modules: `Convolution`

    Parameters:
        model (Module): The PyTorch model for which attribution is to be computed.
        interpolate_mode (Optional[str]): The interpolation mode used by the explainer. Available methods are: `"bilinear"` and `"bicubic"`
        **kwargs: Keyword arguments that are forwarded to the base implementation of the Explainer

    Reference:
        Ramprasaath R. Selvaraju, Michael Cogswell, Abhishek Das, Ramakrishna Vedantam, Devi Parikh, Dhruv Batra. Grad-CAM: Visual Explanations from Deep Networks via Gradient-based Localization.
    """

    SUPPORTED_MODULES = [Convolution]

    def __init__(
        self, model: nn.Module, interpolate_mode: str = "bilinear", **kwargs
    ) -> None:
        super().__init__(model, **kwargs)
        self.interpolate_mode = interpolate_mode

    @property
    def layer(self):
        try:
            return self._layer or find_cam_target_layer(self.model)
        except:
            raise NoCamTargetLayerAndNotTraceableError(
                'You did not set cam target layer and',
                'it does not automatically determined.',
                'Please manually set the cam target layer by:',
                '`Explainer.set_target_layer(layer: nn.Module)`',
                'before attribute.'
            )

    def set_target_layer(self, layer: nn.Module):
        return self.set_kwargs(_layer=layer)

    def attribute(self, inputs: Tensor, targets: Tensor) -> Tensor:
        """
        Computes attributions for the given inputs and targets.

        Args:
            inputs (torch.Tensor): The input data.
            targets (torch.Tensor): The target labels for the inputs.

        Returns:
            torch.Tensor: The result of the explanation.
        """
        forward_args, additional_forward_args = self._extract_forward_args(inputs)
        forward_args = format_into_tuple(forward_args)
        additional_forward_args = format_into_tuple(additional_forward_args)

        assert (
            len(forward_args) == 1
        ), "GradCam for multiple inputs is not supported yet."
        layer = find_cam_target_layer(self.model)
        explainer = LayerGradCam(forward_func=self.model, layer=layer)

        attrs = explainer.attribute(
            forward_args[0],
            target=targets,
            additional_forward_args=additional_forward_args,
            attr_dim_summation=True,
        )
        upsampled = LayerAttribution.interpolate(
            layer_attribution=attrs,
            interpolate_dims=forward_args[0].shape[2:],
            interpolate_mode=self.interpolate_mode,
        )
        return upsampled

    def get_tunables(self) -> Dict[str, Tuple[type, dict]]:
        """
        Provides Tunable parameters for the optimizer

        Tunable parameters:
            `interpolate_mode` (str): Value can be selected of `"bilinear"` and `"bicubic"`
        """
        return {
            "interpolate_mode": (list, {"choices": ["bilinear", "bicubic"]}),
        }



from typing import Optional, Literal

import torchvision.transforms.functional as TF
from torch.nn.modules import Module

from pnpxai.core.detector.types import Convolution
from pnpxai.core._types import Tensor
from pnpxai.explainers.zennit.base import ZennitExplainer
from pnpxai.explainers.zennit.attribution import FullGradient as FullGradAttributor


def _format_pooling_method(method):
    if method == 'abssum':
        return lambda attrs: attrs.abs().sum(1)
    elif method == 'possum':
        return lambda attrs: attrs.clamp(min=0).sum(1)
    else:
        raise ValueError


def _format_interpolate_mode(mode):
    if mode == 'nearest':
        return TF.InterpolationMode.NEAREST
    elif mode == 'nearest_exact':
        return TF.InterpolationMode.NEAREST_EXACT
    elif mode == 'bicubic':
        return TF.InterpolationMode.BICUBIC
    return TF.InterpolationMode.BILINEAR


class FullGrad(ZennitExplainer):
    """
    FullGrad explainer.

    Supported Modules: `Convolution`

    Parameters:
        model (Module): The PyTorch model for which attribution is to be computed.
        pooling_method (Optional[str]): The pooling mode used by the explainer. Available methods are: `"abssum"` (absolute sum) and `"possum"` (positive sum)
        interpolate_mode (Optional[str]): The interpolation mode used by the explainer. Available methods are: `"bilinear"`, `"nearest"`, `"nearest_exact"`, and `"bicubic"`
        n_classes (Optional[int]): The number of classes
        **kwargs: Keyword arguments that are forwarded to the base implementation of the Explainer

    Reference:
        Suraj Srinivas, Francois Fleuret. Full-Gradient Representation for Neural Network Visualization.
    """

    SUPPORTED_MODULES = [Convolution]
    
    def __init__(
        self,
        model: Module,
        pooling_method: Literal['abssum', 'possum']='abssum',
        interpolate_mode: Literal['bilinear', 'nearest', 'nearest_exact', 'bicubic']='blinear',
        n_classes: Optional[int]=None,
    ):
        super().__init__(model=model, n_classes=n_classes)
        self.pooling_method = pooling_method
        self.interpolate_mode = interpolate_mode

    def attribute(self, inputs: Tensor, targets: Tensor):
        """
        Computes attributions for the given inputs and targets.

        Args:
            inputs (torch.Tensor): The input data.
            targets (torch.Tensor): The target labels for the inputs.

        Returns:
            torch.Tensor: The result of the explanation.
        """
        with FullGradAttributor(
            model=self.model,
            pooling_method=_format_pooling_method(self.pooling_method),
            interpolate_mode=_format_interpolate_mode(self.interpolate_mode),
        ) as attributor:
            attrs = attributor(inputs, self._format_targets(targets))
        return attrs

    def get_tunables(self):
        """
            Provides Tunable parameters for the optimizer

            Tunable parameters:
                `pooling_method` (str): Value can be selected of `"abssum"` and `"possum"`
                
                `interpolate_mode` (str): Value can be selected of `"bilinear"`, `"nearest"`, `"nearest_exact"`, and `"bicubic"`
        """
        return {
            'pooling_method': (list, {'choices': ['abssum', 'possum']}),
            'interpolate_mode': (list, {'choices': ['bilinear', 'nearest', 'nearest_exact', 'bicubic']}),
        }



from torch import fx
from pnpxai.core.detector.types import ModuleType


def get_target_module_of(node: fx.Node):
    if node.op != "call_module":
        return
    target_module = node.graph._owning_module
    for i, t in enumerate(node.target.split(".")):
        if not hasattr(target_module, t):
            raise RuntimeError(
                f"Node referenced non-existent target {'.'.join(t[:i])}")
        target_module = getattr(target_module, t)
    return target_module


def find_nearest_user_of(node: fx.Node, module_type: ModuleType):
    for user in node.users:
        module = get_target_module_of(user)
        if isinstance(module, module_type):
            return user
    for user in node.users:
        uuser = find_nearest_user_of(user, module_type)
        if uuser is not None:
            return uuser
    return



from typing import Callable, Optional, Dict, Any


class FunctionSelector:
    """
    A utility class for managing and selecting functions based on keys. This class allows you 
    to define a set of functions associated with specific keys, and provides methods to add, 
    retrieve, and select functions with optional parameters.

    Parameters:
        data (Optional[Dict[str, Callable]], optional):
            A dictionary where the keys are function names (or identifiers) and the values are 
            the corresponding function objects. Defaults to an empty dictionary if not provided.
        default_kwargs (Optional[Dict[str, Any]], optional):
            A dictionary of default keyword arguments to be used when selecting functions. Defaults 
            to an empty dictionary if not provided.

    Attributes:
        choices (List[str]):
            A list of all available function keys.
    
    Methods:
        add(key: str, value: Callable) -> Callable:
            Adds a new function to the selector with the specified key.

        get(key: str) -> Callable:
            Retrieves the function associated with the specified key.

        delete(key: str) -> Optional[Callable]:
            Removes the function associated with the specified key and returns it. If the key does 
            not exist, returns None.

        all() -> List[Callable]:
            Returns a list of all functions currently stored in the selector.

        select(key: str, **kwargs) -> Callable:
            Selects and returns the function associated with the specified key, with the given 
            keyword arguments. Uses default keyword arguments if not overridden.

        get_tunables() -> Dict[str, Any]:
            Returns a dictionary containing tunable parameters for the selector. In this case, it 
            provides the choices of function keys.

    Notes:
        - Ensure that functions added to the selector match the expected signature, as incorrect 
          arguments can lead to runtime errors.
    """
    
    def __init__(
        self,
        data: Optional[Dict[str, Callable]] = None,
        default_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self._data = data or {}
        self._default_kwargs = default_kwargs or {}

    @property
    def choices(self):
        return list(self._data.keys())

    def add(self, key: str, value: Callable):
        self._data[key] = value
        return value

    def get(self, key: str):
        return self._data[key]

    def delete(self, key: str):
        return self._data.pop(key, None)

    def all(self):
        return [self.get(key) for key in self.choices]

    def select(self, key: str, **kwargs):
        fn_type = self.get(key)
        kwargs = {**self._default_kwargs, **kwargs}
        return fn_type(**kwargs)

    def get_tunables(self):
        return {'method': (list, {'choices': self.choices})}



from pnpxai.evaluator.metrics.base import Metric
from pnpxai.evaluator.metrics.mu_fidelity import MuFidelity
from pnpxai.evaluator.metrics.sensitivity import Sensitivity
from pnpxai.evaluator.metrics.complexity import Complexity
from pnpxai.evaluator.metrics.pixel_flipping import (
    PixelFlipping,
    MoRF,
    LeRF,
    AbPC,
)


PIXEL_FLIPPING_METRICS = [
    MoRF,
    LeRF,
    AbPC,
]

AVAILABLE_METRICS = PIXEL_FLIPPING_METRICS



from typing import Optional

import torch
from shap import DeepExplainer
from torch import Tensor
from pnpxai.core.detector.types import Convolution
from pnpxai.explainers.base import Explainer
from torch.nn.modules import Module
from pnpxai.explainers.base import Explainer


class DeepLiftShap(Explainer):
    """
    DeepLiftShap explainer.

    Supported Modules: `Convolution`

    Parameters:
        model (Module): The PyTorch model for which attribution is to be computed.
        background_data: The background dataset used to train the surrogate model.

    Reference:
        Scott M. Lundberg, Su-In Lee, A Unified Approach to Interpreting Model Predictions
    """
    SUPPORTED_MODULES = [Convolution]

    def __init__(
        self,
        model: Module,
        background_data: Tensor,
        dtype: Optional[torch.dtype] = None,
    ):
        super().__init__(model)
        self.background_data = background_data
        self.dtype = dtype or torch.float64

    def attribute(self, inputs: Tensor, targets: Tensor) -> Tensor:
        """
        Computes attributions for the given inputs and targets.

        Args:
            inputs (torch.Tensor): The input data.
            targets (torch.Tensor): The target labels for the inputs.

        Returns:
            torch.Tensor: The result of the explanation.
        """
        _explainer = DeepExplainer(self.model, self.background_data)
        shap_values = _explainer.shap_values(inputs)
        for i in range(len(shap_values)):
            shap_values[i] = shap_values[i][0]
        return torch.tensor(
            shap_values[targets.cpu().tolist()],
            dtype=self.dtype,
            device=self.device,
        )



from typing import Literal, Union, Literal

import torch
import torchvision.transforms.functional as TF
from pnpxai.explainers.utils.base import UtilFunction

BaselineMethod = Literal['zeros', 'invert', 'gaussian_blur', 'token']


class BaselineFunction(UtilFunction):
    """
    A base class for baseline functions used in attribution methods. Baseline functions are 
    used to define a reference or baseline value against which attributions are compared.
    This is typically used to understand the effect of different inputs on the model's predictions.

    Notes:
        - `BaselineFunction` is intended to be subclassed. Concrete baseline functions should 
          inherit from this class and implement the actual baseline logic.
        - Subclasses can override the `__init__` method to accept additional parameters required 
          for their specific baseline operations.
    """    
    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def from_method(cls, method, **kwargs):
        baseline_fn = BASELINE_FUNCTIONS.get(method, None)
        if baseline_fn is None:
            raise ValueError
        return baseline_fn(**kwargs)


class TokenBaselineFunction(BaselineFunction):
    def __init__(self, token_id, **kwargs):
        super().__init__()
        self.token_id = token_id

    def __call__(self, inputs: torch.Tensor):
        return torch.ones_like(inputs, dtype=torch.long) * self.token_id


class ZeroBaselineFunction(BaselineFunction):
    def __init__(self, *args, **kwargs):
        super().__init__()

    def __call__(self, inputs: torch.Tensor):
        return torch.zeros_like(inputs)


class MeanBaselineFunction(BaselineFunction):
    def __init__(self, dim, **kwargs):
        super().__init__()
        self.dim = dim

    def __call__(self, inputs: torch.Tensor):
        size = inputs.size(self.dim)
        repeat = tuple(size if d == self.dim else 1 for d in range(inputs.dim()))
        return torch.mean(inputs, dim=self.dim, keepdim=True).repeat(*repeat)


class InvertBaselineFunction(BaselineFunction):
    def __init__(self, *args, **kwargs):
        super().__init__()

    def __call__(self, inputs: torch.Tensor):
        return TF.invert(inputs)


class GaussianBlurBaselineFunction(BaselineFunction):
    def __init__(
        self,
        kernel_size_x: int = 3,
        kernel_size_y: int = 3,
        sigma_x: float = .5,
        sigma_y: float = .5,
        **kwargs
    ):
        super().__init__()
        self.kernel_size_x = kernel_size_x
        self.kernel_size_y = kernel_size_y
        self.sigma_x = sigma_x
        self.sigma_y = sigma_y

    def __call__(self, inputs: torch.Tensor):
        return TF.gaussian_blur(
            inputs,
            kernel_size=[self.kernel_size_x, self.kernel_size_y],
            sigma=[self.sigma_x, self.sigma_y],
        )

    def get_tunables(self):
        return {
            'kernel_size_x': (int, {'low': 1, 'high': 11, 'step': 2}),
            'kernel_size_y': (int, {'low': 1, 'high': 11, 'step': 2}),
            'sigma_x': (float, {'low': .05, 'high': 2., 'step': .05}),
            'sigma_y': (float, {'low': .05, 'high': 2., 'step': .05}),
        }


BaselineMethodOrFunction = Union[BaselineMethod, BaselineFunction]

BASELINE_FUNCTIONS_FOR_IMAGE = {
    'zeros': ZeroBaselineFunction,
    'mean': MeanBaselineFunction,
    'invert': InvertBaselineFunction,
    'gaussian_blur': GaussianBlurBaselineFunction,
}

BASELINE_FUNCTIONS_FOR_TEXT = {
    'token': TokenBaselineFunction,
}

BASELINE_FUNCTIONS_FOR_TIME_SERIES = {
    'zeros': ZeroBaselineFunction,
    'mean': MeanBaselineFunction,
}

BASELINE_FUNCTIONS = {
    **BASELINE_FUNCTIONS_FOR_IMAGE,
    **BASELINE_FUNCTIONS_FOR_TEXT,
    **BASELINE_FUNCTIONS_FOR_TIME_SERIES,
}


from pnpxai.explainers.utils.utils import (
    find_cam_target_layer,
    get_default_feature_mask_fn,
    captum_wrap_model_input,
    _format_to_tuple,
)
from pnpxai.explainers.utils.base import UtilFunction
from pnpxai.explainers.utils.baselines import BaselineFunction
from pnpxai.explainers.utils.feature_masks import FeatureMaskFunction
from pnpxai.explainers.utils.function_selectors import FunctionSelector
from pnpxai.explainers.utils.postprocess import (
    PoolingFunction,
    NormalizationFunction,
    PostProcessor,
)



EVALUATION_METRIC_AUTO_KWARGS = {}

EVALUATION_METRIC_REVERSE_SORT = {
    "MuFidelity": True,
    "Complexity": False,
    "Sensitivity": False
}

EVALUATION_METRIC_SORT_PRIORITY = [
    "MuFidelity",
    "Sensitivity",
    "Complexity"
]


from typing import Optional

from torch.nn.modules import Module
from zennit.core import BasicHook, Stabilizer
from zennit.composites import LayerMapComposite
from zennit.rules import NoMod
from zennit.types import Linear

from pnpxai.explainers.lrp import LRPBase, canonizers_base
from pnpxai.explainers.types import (
    ForwardArgumentExtractor,
    TargetLayer,
)


class GuidedBackpropRule(BasicHook):
    def __init__(self, stabilizer=1e-6, zero_params=None) -> None:
        stabilizer_fn = Stabilizer.ensure(stabilizer)
        super().__init__(
            input_modifiers=[lambda input: input],
            param_modifiers=[NoMod(zero_params=zero_params)],
            output_modifiers=[lambda output: output.clamp(min=0)],
            gradient_mapper=(lambda out_grad, outputs: out_grad.clamp(min=0) / stabilizer_fn(outputs[0])),
            reducer=(lambda inputs, gradients: inputs[0] * gradients[0]),
        )


class GuidedBackprop(LRPBase):
    def __init__(
        self,
        model: Module,
        stabilizer: float=1e-6,
        forward_arg_extractor: Optional[ForwardArgumentExtractor]=None,
        additional_forward_arg_extractor: Optional[ForwardArgumentExtractor]=None,
        layer: Optional[TargetLayer]=None,
        n_classes: Optional[int]=None,
    ) -> None:
        self.stabilizer = stabilizer
        super().__init__(
            model,
            self._composite,
            forward_arg_extractor,
            additional_forward_arg_extractor,
            layer,
            n_classes
        )

    @property
    def _layer_map(self):
        return [(Linear, GuidedBackpropRule(stabilizer=self.stabilizer))]

    @property
    def _canonizers(self):
        return canonizers_base()

    @property
    def _composite(self):
        return LayerMapComposite(
            layer_map=self._layer_map,
            canonizers=self._canonizers,
        )